#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
from ase import Atoms as aseAtoms

from nomad.units import ureg
from nomad.parsing.file_parser import Quantity, TextParser

from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.method import Method
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.calculation import Calculation
from nomad.datamodel.metainfo.workflow import Workflow, Elastic, StrainDiagrams

from .metainfo.elastic import x_elastic_section_fitting_parameters


class InfoParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'order', r'\s*Order of elastic constants\s*=\s*([0-9]+)', repeats=False,
                dtype=int),
            Quantity(
                'calculation_method', r'\s*Method of calculation\s*=\s*([-a-zA-Z]+)\s*',
                repeats=False),
            Quantity(
                'code_name', r'\s*DFT code name\s*=\s*([-a-zA-Z]+)', repeats=False),
            Quantity(
                'space_group_number', r'\s*Space-group number\s*=\s*([0-9]+)', repeats=False),
            Quantity(
                'equilibrium_volume', r'\s*Volume of equilibrium unit cell\s*=\s*([0-9.]+)\s*',
                unit='angstrom ** 3'),
            Quantity(
                'max_strain', r'\s*Maximum Lagrangian strain\s*=\s*([0-9.]+)', repeats=False),
            Quantity(
                'n_strains', r'\s*Number of distorted structures\s*=\s*([0-9]+)', repeats=False)]


class StructureParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        def get_sym_pos(val):
            val = val.strip().replace('\n', '').split()
            sym = []
            pos = []
            for i in range(0, len(val), 4):
                sym.append(val[i + 3].strip())
                pos.append([float(val[j]) for j in range(i, i + 3)])
            sym_pos = dict(symbols=sym, positions=pos)
            return sym_pos

        self._quantities = [
            Quantity(
                'cellpar', r'a\s*b\s*c\n([\d\.\s]+)\n\s*alpha\s*beta\s*gamma\n([\d\.\s]+)\n+',
                repeats=False),
            Quantity(
                'sym_pos', r'Atom positions:\n\n([\s\d\.A-Za-z]+)\n\n',
                str_operation=get_sym_pos, repeats=False, convert=False)]


class DistortedParametersParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        self._quantities = [Quantity(
            'deformation', r'Lagrangian strain\s*=\s*\(([eta\s\d\.,]+)\)',
            str_operation=lambda x: x.replace(',', '').split(), repeats=True, dtype=str)]


class FitParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):

        def split_eta_val(val):
            order, val = val.strip().split(' order fit.')
            val = [float(v) for v in val.strip().split()]
            return order, val[0::2], val[1::2]

        self._quantities = [Quantity(
            'fit', r'(\w+ order fit\.\n[\d.\s\neE\-\+]+)\n', repeats=True, convert=False,
            str_operation=split_eta_val)]


class ElasticConstant2Parser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'voigt', r'Symmetry[\s\S]+\n\s*\n([C\d\s\n\(\)\-\+\/\*]+)\n',
                shape=(6, 6), dtype=str, repeats=False),
            Quantity(
                'elastic_constant', r'Elastic constant[\s\S]+in GPa\s*:\s*\n\n([\-\d\.\s\n]+)\n',
                shape=(6, 6), dtype=float, unit='GPa', repeats=False),
            Quantity(
                'compliance', r'Elastic compliance[\s\S]+in 1/GPa\s*:\s*\n\n([\-\d\.\s\n]+)\n',
                shape=(6, 6), dtype=float, unit='1/GPa', repeats=False)]

        def str_to_modulus(val_in):
            val_in = val_in.strip().split()
            key = val_in[0]
            unit = val_in[-1] if len(val_in) == 3 else None
            val = float(val_in[1])
            val = val * ureg.GPa if unit is not None else val
            return key, val

        self._quantities.append(Quantity(
            'modulus', r',\s*(\w+)\s*=\s*([\-\+\w\. ]+?)\n', str_operation=str_to_modulus,
            repeats=True))

        self._quantities.append(Quantity(
            'eigenvalues',
            r'Eigenvalues of elastic constant \(stiffness\) matrix:\s*\n+([\-\d\.\n\s]+)\n',
            unit='GPa', repeats=False))


class ElasticConstant3Parser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        def arrange_matrix(val):
            val = val.strip().split('\n')
            matrix = [v.strip().split() for v in val if v.strip()]
            matrix = np.array(matrix).reshape((12, 18))
            arranged = []
            for i in range(2):
                for j in range(3):
                    arranged.append(
                        matrix[i * 6: (i + 1) * 6, j * 6: (j + 1) * 6].tolist())
            return arranged

        self._quantities = [
            Quantity(
                'elastic_constant', r'\%\s*\n([\s0-6A-L]*)[\n\s\%1-6\-ij]*([\s0-6A-L]*)\n',
                str_operation=arrange_matrix, dtype=str, repeats=False, convert=False),
            Quantity(
                'cijk', r'(C\d\d\d)\s*=\s*([\-\d\.]+)\s*GPa', repeats=True, convert=False)]


class ElasticParser:
    def __init__(self):
        self._mainfile = None
        self.logger = None
        self._deform_dirs = None
        self._deform_dir_prefix = 'Dst'
        self._dirs = []
        self.info = InfoParser()
        self.structure = StructureParser()
        self.distorted_parameters = DistortedParametersParser()
        self.fit = FitParser()
        self.elastic_constant_2 = ElasticConstant2Parser()
        self.elastic_constant_3 = ElasticConstant3Parser()

    @property
    def deformation_dirs(self):
        if self._deform_dirs is None:
            self._deform_dirs = [
                os.path.join(self.maindir, d) for d in self._dirs if d.startswith(self._deform_dir_prefix)]

        return self._deform_dirs

    def get_elastic_files(self, filename, extension, dirname=None):
        dirs = self._dirs if dirname is None else os.listdir(dirname)
        dirname = self.maindir if dirname is None else dirname
        filenames = [d for d in dirs if filename in d and d.endswith(extension)]
        if len(filenames) > 1:
            filenames = [d for d in filenames if d.startswith(filename)]
        if len(filenames) == 0:
            return
        return os.path.join(dirname, filenames[0])

    def get_references_to_calculations(self):
        def output_file(dirname):
            code = self.info.get('code_name', '').lower()
            if code == 'exciting':
                return os.path.join(dirname, 'INFO.OUT')
            elif code == 'wien':
                return os.path.join(dirname, '%s_Converged.scf' % os.path.basename(dirname))
            elif code == 'quantum':
                return os.path.join(dirname, '%s.out' % os.path.basename(dirname))
            else:
                return None

        references = []
        for deform_dir in self.deformation_dirs:
            sub_dirs = os.listdir(deform_dir)
            for sub_dir in sub_dirs:
                calc_dir = os.path.join(deform_dir, sub_dir)
                out_file = output_file(calc_dir)
                if out_file is not None and os.path.isfile(out_file):
                    references.append(out_file)

        return references

    def get_structure_info(self):
        path = os.path.join(self.maindir, 'sgroup.out')
        if not os.path.isfile(path):
            return

        self.structure.mainfile = path

        cellpar = self.structure.get('cellpar', None)
        sym_pos = self.structure.get('sym_pos', {})

        sym = sym_pos.get('symbols', None)
        pos = sym_pos.get('positions', None)

        if cellpar is None or sym is None or pos is None:
            return

        structure = aseAtoms(cell=cellpar, scaled_positions=pos, symbols=sym, pbc=True)

        positions = structure.get_positions()
        positions = positions * ureg.angstrom
        cell = structure.get_cell()
        cell = cell * ureg.angstrom

        return sym, positions, cell

    def get_strain_energy(self):
        strains, energies = [], []

        for deform_dir in self.deformation_dirs:
            filenames = [d for d in os.listdir(deform_dir) if d.endswith('Energy.dat')]
            if not filenames:
                continue

            path = os.path.join(deform_dir, filenames[-1])
            data = np.loadtxt(path).T
            strains.append(list(data[0]))
            # the peculiarity of the x_elastic_strain_diagram_values metainfo that it does
            # not have the energy unit
            energies.append((data[1] * ureg.hartree).to('J').magnitude)
        if len(np.shape(energies)) != 2:
            strains, energies = [], []
        return strains, energies

    def get_strain_stress(self):
        strains = {'Lagrangian-stress': [], 'Physical-stress': []}
        stresses = {'Lagrangian-stress': [], 'Physical-stress': []}

        for deform_dir in self.deformation_dirs:
            filenames = [d for d in os.listdir(deform_dir) if d.endswith('stress.dat')]

            for filename in filenames:
                path = os.path.join(deform_dir, filename)
                if not os.path.isfile(path):
                    continue

                with open(path) as f:
                    lines = f.readlines()

                strain, stress = [], []
                for line in lines:
                    val = line.strip().split()
                    if not val[0].strip().replace('.', '').isdecimal():
                        continue

                    strain.append(float(val[0]))
                    stress.append([float(v) for v in val[1:7]])

                stype = filename.rstrip('.dat').split('_')[-1]
                strains[stype].append(strain)
                stresses[stype].append(stress)

        return strains, stresses

    def get_deformation_types(self):
        path = os.path.join(self.maindir, 'Distorted_Parameters')
        self.distorted_parameters.mainfile = path
        return self.distorted_parameters.get('deformation')

    def _get_fit(self, path_dir, file_ext):
        path_dir = os.path.join(self.maindir, path_dir)

        if not os.path.isdir(path_dir):
            return

        paths = [p for p in os.listdir(path_dir) if p.endswith(file_ext)]
        paths.sort()

        if not paths:
            return

        eta, val = {}, {}
        for path in paths:
            self.fit.mainfile = os.path.join(path_dir, path)
            fit_results = self.fit.get('fit', [])
            for result in fit_results:
                eta.setdefault(result[0], [])
                val.setdefault(result[0], [])
                eta[result[0]].append(result[1])
                val[result[0]].append(result[2])

        return eta, val

    def get_energy_fit(self):
        energy_fit = dict()

        for file_ext in ['d2E.dat', 'd3E.dat', 'ddE.dat']:
            result = self._get_fit('Energy-vs-Strain', file_ext)
            if result is None:
                continue

            result = list(result)
            result[1] = {
                key: (val * ureg.GPa).to('Pa').magnitude for key, val in result[1].items()}
            energy_fit['d2e'] = result

        result = self._get_fit('Energy-vs-Strain', 'CVe.dat')
        if result is not None:
            result = list(result)
            result[1] = {
                key: (val * ureg.hartree).to('J').magnitude for key, val in result[1].items()}
            energy_fit['cross-validation'] = result

        return energy_fit

    def get_stress_fit(self):
        stress_fit = dict()
        stress_fit['dtn'] = [[]] * 6
        stress_fit['cross-validation'] = [[]] * 6

        for strain_index in range(1, 7):
            result = self._get_fit('Stress-vs-Strain', '%d_dS.dat' % strain_index)
            if result is not None:
                result[1] = {key: val * ureg.GPa for key, val in result[1].items()}
                stress_fit['dtn'][strain_index - 1] = result

            result = self._get_fit('Stress-vs-Strain', '%d_CVe.dat' % strain_index)
            if result is not None:
                result[1] = {key: val * ureg.hartree for key, val in result[1].items()}
                stress_fit['cross-validation'][strain_index - 1] = result

        return stress_fit

    def get_input(self):
        paths = os.listdir(self.maindir)
        path = None
        order = self.info.get('order', 2)
        for p in paths:
            if 'ElaStic_' in p and p.endswith('.in') and str(order) in p:
                path = p
                break

        if path is None:
            return

        calc_method = self.info.get('calculation_method')

        eta_ec = []
        fit_ec = []

        def _is_number(var):
            try:
                float(var)
                return True
            except Exception:
                return False

        path = os.path.join(self.maindir, path)

        with open(path) as f:
            while True:
                line = f.readline()
                if not line:
                    break

                if calc_method.lower() == 'energy':
                    _, eta, fit = line.strip().split()
                    eta_ec.append(float(eta))
                    fit_ec.append(int(fit))

                elif calc_method.lower() == 'stress':
                    val = line.strip().split()
                    if not _is_number(val[0]):
                        eta_ec.append([float(val[i + 1]) for i in range(6)])
                    else:
                        fit_ec.append([int(val[i]) for i in range(6)])

                else:
                    pass

        return eta_ec, fit_ec

    def get_elastic_constants_order2(self):
        path = self.get_elastic_files('ElaStic_2nd', 'out')
        self.elastic_constant_2.mainfile = path

        matrices = dict()
        for key in ['voigt', 'elastic_constant', 'compliance']:
            val = self.elastic_constant_2.get(key, None)
            if val is not None:
                matrices[key] = val

        moduli = dict()
        for modulus in self.elastic_constant_2.get('modulus', []):
            moduli[modulus[0]] = modulus[1]

        eigenvalues = self.elastic_constant_2.get('eigenvalues')

        return matrices, moduli, eigenvalues

    def get_elastic_constants_order3(self):
        path = self.get_elastic_files('ElaStic_3rd.out', 'out')
        self.elastic_constant_3.mainfile = path

        elastic_constant_str = self.elastic_constant_3.get('elastic_constant')
        if elastic_constant_str is None:
            return

        cijk = dict()
        for element in self.elastic_constant_3.get('cijk', []):
            cijk[element[0]] = float(element[1])

        # formulas for the coefficients
        coeff_A = cijk.get('C111', 0) + cijk.get('C112', 0) - cijk.get('C222', 0)
        coeff_B = -(cijk.get('C115', 0) + 3 * cijk.get('C125', 0)) / 2
        coeff_C = (cijk.get('C114', 0) + 3 * cijk.get('C124', 0)) / 2
        coeff_D = -(2 * cijk.get('C111', 0) + cijk.get('C112', 0) - 3 * cijk.get('C222', 0)) / 4
        coeff_E = -cijk.get('C114', 0) - 2 * cijk.get('C124', 0)
        coeff_F = -cijk.get('C115', 0) - 2 * cijk.get('C125', 0)
        coeff_G = -(cijk.get('C115', 0) - cijk.get('C125', 0)) / 2
        coeff_H = (cijk.get('C114', 0) - cijk.get('C124', 0)) / 2
        coeff_I = (2 * cijk.get('C111', 0) - cijk.get('C112', 0) - cijk.get('C222', 0)) / 4
        coeff_J = (cijk.get('C113', 0) - cijk.get('C123', 0)) / 2
        coeff_K = -(cijk.get('C144', 0) - cijk.get('C155', 0)) / 2

        space_group_number = self.info.get('space_group_number')

        if space_group_number <= 148:  # rhombohedral II
            coefficients = dict(
                A=coeff_A, B=coeff_B, C=coeff_C, D=coeff_D, E=coeff_E, F=coeff_F, G=coeff_G,
                H=coeff_H, I=coeff_I, J=coeff_J, K=coeff_K)
        elif space_group_number <= 167:  # rhombohedral I
            coefficients = dict(
                A=coeff_A, B=coeff_C, C=coeff_D, D=coeff_E, E=coeff_H, F=coeff_I, G=coeff_J,
                H=coeff_K)
        elif space_group_number <= 176:  # hexagonal II
            coefficients = dict(
                A=coeff_A, B=coeff_D, C=coeff_I, D=coeff_J, E=coeff_K)
        elif space_group_number <= 194:  # hexagonal I
            coefficients = dict(
                A=coeff_A, B=coeff_D, C=coeff_I, D=coeff_J, E=coeff_K)
        else:
            coefficients = dict()

        # assign values to the elastic constant matrix from the independent ec and coefficients
        elastic_constant = np.zeros((6, 6, 6))
        for i in range(6):
            for j in range(6):
                for k in range(6):
                    val = elastic_constant_str[i][j][k]
                    if val == '0':
                        elastic_constant[i][j][k] = 0
                    elif val.isdigit():
                        elastic_constant[i][j][k] = cijk['C%s' % val]
                    else:
                        elastic_constant[i][j][k] = coefficients.get(val, 0)

        return elastic_constant

    def parse_strain(self):
        sec_elastic = self.archive.workflow[0].elastic
        method = self.info['calculation_method'].lower()

        n_strains = self.info['n_strains']
        poly_fit_2 = int((n_strains - 1) / 2)
        poly_fit = {
            '2nd': poly_fit_2, '3rd': poly_fit_2 - 1, '4th': poly_fit_2 - 1,
            '5th': poly_fit_2 - 2, '6th': poly_fit_2 - 2, '7th': poly_fit_2 - 3}

        if method == 'energy':
            strain, energy = self.get_strain_energy()
            if not strain:
                self.logger.warn('Error getting strain and energy data')
                return

            sec_strain_diagram = sec_elastic.m_create(StrainDiagrams)
            sec_strain_diagram.type = 'energy'
            sec_strain_diagram.n_eta = len(strain[0])
            sec_strain_diagram.eta = strain
            sec_strain_diagram.value = energy

            energy_fit = self.get_energy_fit()
            if not energy_fit:
                self.logger.warn('Error getting energy fit data')
                return

            for diagram_type in ['cross-validation', 'd2e']:
                for fit_order in energy_fit[diagram_type][0].keys():
                    sec_strain_diagram = sec_elastic.m_create(StrainDiagrams)
                    sec_strain_diagram.type = diagram_type
                    sec_strain_diagram.polynomial_fit_order = int(fit_order[:-2])
                    sec_strain_diagram.n_eta = poly_fit.get(fit_order, None)
                    sec_strain_diagram.eta = energy_fit[diagram_type][0][fit_order]
                    sec_strain_diagram.value = energy_fit[diagram_type][1][fit_order]

        elif method == 'stress':
            strain, stress = self.get_strain_stress()
            for diagram_type in ['Lagrangian-stress', 'Physical-stress']:
                strain_i = strain[diagram_type]
                if not strain_i:
                    continue
                stress_i = np.transpose(np.array(stress[diagram_type]), axes=(2, 0, 1))

                for si in range(6):
                    sec_strain_diagram = sec_elastic.m_create(StrainDiagrams)
                    sec_strain_diagram.type = diagram_type
                    sec_strain_diagram.stress_voigt_component = si + 1
                    sec_strain_diagram.n_eta = len(strain_i[0])
                    sec_strain_diagram.eta = strain_i
                    sec_strain_diagram.value = stress_i[si]

            stress_fit = self.get_stress_fit()
            for diagram_type in ['cross-validation', 'dtn']:
                for si in range(6):
                    if len(stress_fit[diagram_type][si]) == 0:
                        continue
                    for fit_order in stress_fit[diagram_type][si][0].keys():
                        sec_strain_diagram = sec_elastic.m_create(StrainDiagrams)
                        sec_strain_diagram.type = diagram_type
                        sec_strain_diagram.stress_voigt_component = si + 1
                        sec_strain_diagram.polynomial_fit_order = int(fit_order[:-2])
                        sec_strain_diagram.n_eta = poly_fit.get(fit_order, None)
                        sec_strain_diagram.eta = stress_fit[diagram_type][si][0][fit_order]
                        sec_strain_diagram.value = np.array(stress_fit[diagram_type][si][1][fit_order])

    def parse_elastic_constant(self):
        sec_elastic = self.archive.workflow[0].elastic

        order = self.info['order']

        if order == 2:
            matrices, moduli, eigenvalues = self.get_elastic_constants_order2()
            sec_elastic.elastic_constants_notation_matrix_second_order = matrices.get('voigt')
            sec_elastic.elastic_constants_matrix_second_order = matrices.get('elastic_constant')
            sec_elastic.compliance_matrix_second_order = matrices.get('compliance')

            sec_elastic.bulk_modulus_voigt = moduli.get('B_V', moduli.get('K_V'))
            sec_elastic.shear_modulus_voigt = moduli.get('G_V')

            sec_elastic.bulk_modulus_reuss = moduli.get('B_R', moduli.get('K_R'))
            sec_elastic.shear_modulus_reuss = moduli.get('G_R')

            sec_elastic.bulk_modulus_hill = moduli.get('B_H', moduli.get('K_H'))
            sec_elastic.shear_modulus_hill = moduli.get('G_H')

            sec_elastic.young_modulus_voigt = moduli.get('E_V')
            sec_elastic.poisson_ratio_voigt = moduli.get('nu_V')
            sec_elastic.young_modulus_reuss = moduli.get('E_R')
            sec_elastic.poisson_ratio_reuss = moduli.get('nu_R')
            sec_elastic.young_modulus_hill = moduli.get('E_H')
            sec_elastic.poisson_ratio_hill = moduli.get('nu_H')

            sec_elastic.eigenvalues_elastic = eigenvalues

        elif order == 3:
            elastic_constant = self.get_elastic_constants_order3()
            if elastic_constant is not None:
                sec_elastic.elastic_constants_matrix_third_order = elastic_constant * ureg.GPa

    def init_parser(self):
        self._deform_dirs = None
        self.maindir = os.path.dirname(self.filepath)
        self.info.mainfile = self.filepath
        self.info.logger = self.logger
        self.structure.logger = self.logger
        self.distorted_parameters.logger = self.logger
        self.fit.logger = self.logger
        self.elastic_constant_2.logger = self.logger
        self.elastic_constant_3.logger = self.logger
        self._dirs = os.listdir(self.maindir)

    def reuse_parser(self, parser):
        self.info.quantities = parser.info.quantities
        self.structure.quantities = parser.structure.quantities
        self.distorted_parameters.quantities = parser.distorted_parameters.quantities
        self.fit.quantities = parser.fit.quantities
        self.elastic_constant_2.quantities = parser.elastic_constant_2.quantities
        self.elastic_constant_3.quantities = parser.elastic_constant_3.quantities

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program = Program(name='elastic', version='1.0')

        sec_system = sec_run.m_create(System)

        symbols_positions_cell = self.get_structure_info()
        volume = self.info['equilibrium_volume']

        sec_atoms = sec_system.m_create(Atoms)
        if symbols_positions_cell is not None:
            sec_atoms.labels = symbols_positions_cell[0]
            sec_atoms.positions = symbols_positions_cell[1]
            sec_atoms.lattice_vectors = symbols_positions_cell[2]
        sec_atoms.periodic = [True, True, True]
        sec_system.x_elastic_space_group_number = self.info['space_group_number']
        sec_system.x_elastic_unit_cell_volume = volume

        sec_method = sec_run.m_create(Method)

        sec_scc = sec_run.m_create(Calculation)
        sec_scc.calculations_path = self.get_references_to_calculations()

        fit_input = self.get_input()
        if fit_input is not None:
            sec_fit_par = sec_method.m_create(x_elastic_section_fitting_parameters)
            sec_fit_par.x_elastic_fitting_parameters_eta = fit_input[0]
            sec_fit_par.x_elastic_fitting_parameters_polynomial_order = fit_input[1]

        sec_scc.method_ref = sec_method
        sec_scc.system_ref = sec_system

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.workflow_type = 'elastic'
        sec_elastic = sec_workflow.m_create(Elastic)
        sec_elastic.energy_stress_calculator = self.info['code_name']
        sec_elastic.calculation_method = self.info['calculation_method'].lower()
        sec_elastic.elastic_constants_order = self.info['order']
        sec_elastic.strain_maximum = self.info['max_strain']
        sec_elastic.n_strains = self.info['n_strains']

        deformation_types = self.get_deformation_types()
        sec_elastic.n_deformations = len(self.deformation_dirs)
        sec_elastic.deformation_types = deformation_types

        self.parse_strain()
        self.parse_elastic_constant()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Parser for creating artificial test, brenchmark, and demonstration data.
'''

import json
import os.path
import random
from ase.data import chemical_symbols
import numpy
import sys
import time
import os
import signal

from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.method import Electronic, XCFunctional, DFT
from nomad.datamodel.metainfo.simulation.system import Atoms

from .parser import Parser, MatchingParser


class EmptyParser(MatchingParser):
    '''
    Implementation that produces an empty code_run
    '''
    name = "parsers/empty"

    def parse(self, mainfile: str, archive: EntryArchive, logger=None, child_archives=None) -> None:
        run = archive.m_create(Run)
        run.program = Program(name=self.code_name)


class TemplateParser(Parser):
    '''
    A parser that generates data based on a template given via the
    mainfile. The template is basically some archive json. Only
    '''
    name = 'parsers/template'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code_name = 'Template'

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return filename.endswith('template.json')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None, child_archives=None) -> None:
        # tell tests about received logger
        if logger is not None:
            logger.debug('received logger')

        template_json = json.load(open(mainfile, 'r'))
        loaded_archive = EntryArchive.m_from_dict(template_json)
        archive.m_add_sub_section(EntryArchive.run, loaded_archive.run[0])
        archive.m_add_sub_section(EntryArchive.workflow, loaded_archive.workflow[0])

        if 'warning' in mainfile:
            logger.warn('a test warning.')

        logger.debug('a test log entry')


class ChaosParser(Parser):
    '''
    Parser that emulates typical error situations. Files can contain a json string (or
    object with key `chaos`) with one of the following string values:
    - exit
    - deadlock
    - consume_ram
    - exception
    - segfault
    - random
    '''
    name = 'parsers/chaos'

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return filename.endswith('chaos.json')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None, child_archives=None) -> None:
        chaos_json = json.load(open(mainfile, 'r'))
        if isinstance(chaos_json, str):
            chaos = chaos_json
        elif isinstance(chaos_json, dict):
            chaos = chaos_json.get('chaos', None)
        else:
            chaos = None

        if chaos == 'random':
            chaos = random.choice(['exit', 'deadlock', 'consume_ram', 'exception', 'segfault'])

        if chaos == 'exit':
            sys.exit(1)
        elif chaos == 'deadlock':
            while True:
                time.sleep(1)
        elif chaos == 'consume_ram':
            data = []
            i = 0
            while True:
                data.append('a' * 10**6)
                i += 1
                logger.info('ate %d mb' % i)
        elif chaos == 'exception':
            raise Exception('Some chaos happened, muhuha...')
        elif chaos == 'segfault':
            os.kill(os.getpid(), signal.SIGSEGV)

        raise Exception('Unknown chaos')


class GenerateRandomParser(TemplateParser):
    name = 'parsers/random'

    basis_set_types = [
        'Numeric AOs', 'Gaussians', '(L)APW+lo', 'Plane waves',
        'Real-space grid', 'Local-orbital minimum-basis']

    electronic_structure_methods = [
        'DFT', 'DFT+U', 'full-CI', 'CIS', 'CISD'
        'CCS', 'CCS(D)', 'CCSD', 'CCSD(T)', 'CCSDT(Q)', 'MP2', 'MP3', 'MP4', 'MP5', 'MP6',
        'G0W0', 'scGW', 'LDA', 'hybrid', 'CASPT2', 'MRCIS', 'MRCISD', 'RAS-CI']

    XC_functional_names = [
        'LDA_X', 'LDA_C', 'GGA_X', 'GGA_C', 'HYB_GGA_XC', 'MGGA_X', 'MGGA_C', 'HYB_MGGA_XC']

    low_numbers = [1, 1, 2, 2, 2, 2, 2, 3, 3, 4]

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return os.path.basename(filename).startswith('random_')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None, child_archives=None) -> None:
        file_dir = os.path.dirname(os.path.abspath(__file__))
        relative_template_file = "random_template.json"
        template_file = os.path.normpath(os.path.join(file_dir, relative_template_file))

        super().parse(template_file, archive, logger)

        seed = int(os.path.basename(mainfile).split('_')[1])
        random.seed(seed)
        numpy.random.seed(seed)

        run = archive.run[0]

        for system in run.system:
            atoms = []
            atom_positions = []
            # different atoms
            for _ in range(0, random.choice(GenerateRandomParser.low_numbers)):
                # ase data starts with a place holder value X, we don't want that
                atom = random.choice(chemical_symbols[1:])
                # atoms of the same number
                for _ in range(0, random.choice(GenerateRandomParser.low_numbers)):
                    atoms.append(atom)
                    atom_positions.append([.0, .0, .0])

            system.atoms = Atoms(labels=atoms, positions=atom_positions)

        run.program_basis_set_type = random.choice(GenerateRandomParser.basis_set_types)

        for method in run.method:
            method.electronic = Electronic(method=random.choice(GenerateRandomParser.electronic_structure_methods))
            method.dft = DFT(xc_functional=XCFunctional(name=random.choice(GenerateRandomParser.XC_functional_names)))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import logging
from ase.data import chemical_symbols

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity

from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.method import (
    Method, AtomParameters, DFT, XCFunctional, Functional, Electronic, BasisSet,
    BasisSetAtomCentered)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, ScfIteration, Energy, EnergyEntry, Forces, ForcesEntry, BandEnergies,
    Charges, ChargesValue, Multipoles, MultipolesEntry
)


re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[Ee][-+]\d+)?'


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        key_val_quantity = Quantity('key_val', r'([A-Z]+) *= *(\S+)', repeats=True)

        def to_gradient(val_in):
            gradient = []
            for val in val_in.strip().splitlines():
                val = val.strip()
                if val and val[0].isdecimal():
                    gradient.append(val.split()[-3:])
            return np.array(gradient, dtype=np.float64) * ureg.hartree / ureg.bohr

        def to_charges(val_in):
            charges = []
            for val in val_in.strip().splitlines():
                val = val.strip()
                if val and val[0].isdecimal():
                    charges.append(val.split()[2:])
            return np.transpose(charges)

        calc_quantities = [
            Quantity(
                'scf',
                r'(SCF CALCULATION[\s\S]+?)END OF .+? CALCULATION',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        rf'ITER.+?TOTAL ENERGY.+([\s\S]+?){re_n} *{re_n}',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'iter',
                                rf'({re_f} +{re_f}).+',
                                dtype=np.dtype(np.float64), repeats=True
                            )
                        ])
                    ),
                    Quantity(
                        'energy_total',
                        rf'FINAL .+? ENERGY IS +({re_f})', unit=ureg.hartree
                    ),
                    Quantity(
                        'converged',
                        r'(DENSITY CONVERGED)', str_operation=lambda x: True
                    ),
                    Quantity(
                        'eigenvectors',
                        r'EIGENVECTORS\s+\-+([\s\S]+?)(?:\.{6}|\-{10})',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'eigenvalues',
                                rf'{re_n} +({re_f}.+)', str_operation=lambda x: x.split(), repeats=True)])
                    ),

                ])
            ),
            Quantity(
                'coordinates',
                rf'((?:COORDINATES OF ALL ATOMS ARE \(ANGS\)\s+ATOM +|ATOM +ATOMIC +COORDINATES \(BOHR\)\s+)'
                rf'CHARGE +X +Y +Z\s+\-*[\s\S]+?)INTERNUCLEAR',
                sub_parser=TextParser(quantities=[
                    Quantity('unit', r'\((\S+)\)'),
                    Quantity('label_charge', rf'([A-Z]+[a-z]*) +({re_f})', repeats=True),
                    Quantity('position', rf'({re_f} +{re_f} +{re_f}) *{re_n}', dtype=np.dtype(np.float64), repeats=True)
                ])
            ),
            Quantity(
                'gradient',
                r'ATOM +ZNUC + DE/DX +DE/DY +DE/DZ\s+\-+([\s\S]+?)MAXIMUM',
                str_operation=to_gradient
            ),
            Quantity(
                'properties',
                r'(?:properties for the|PROPERTY VALUES FOR THE|PROPERTIES FOR THE)([\s\S]+?)END OF PROPERTY EVALUATION',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_components',
                        r'ENERGY COMPONENTS\s+\-+([\s\S]+?VIRIAL RATIO.+)',
                        str_operation=lambda x: [v.strip().split('=') for v in x.strip().splitlines() if '=' in v]
                    ),
                    Quantity(
                        'population_analysis',
                        r'MULLIKEN AND LOWDIN POPULATION ANALYSES\s+\-+([\s\S]+?(?:\-{30}|\Z))',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'atomic',
                                rf'TOTAL MULLIKEN AND LOWDIN ATOMIC POPULATIONS\s+'
                                rf'ATOM +MULL.POP\. +CHARGE +LOW\.POP\. +CHARGE\s+([\s\S]+?)(?:{re_n} +[A-Z]|\Z)',
                                str_operation=to_charges
                            ),
                            Quantity(
                                'spherical_harmonics',
                                rf'SPHERICAL HARMONIC POPULATIONS\s+ATOM.+\s+([\s\S]+?)(?:{re_n} +[A-Z]|\Z)',
                                str_operation=to_charges, repeats=True
                            ),
                            Quantity(
                                'atomic_orbitals',
                                r'POPULATIONS IN EACH AO +\-+\s+MULLIKEN +LOWDIN\s+([\s\S]+?)\-\-\-\-\-',
                                str_operation=to_charges
                            ),
                            Quantity(
                                'spins',
                                r'ATOMIC SPIN POPULATION \(ALPHA MINUS BETA\)\s+ATOM.+\s+([\s\S]+?)\-\-\-\-\-',
                                str_operation=to_charges
                            )
                        ])
                    ),
                    Quantity(
                        'electrostatic_moments',
                        r'ELECTROSTATIC MOMENTS\s+\-+([\s\S]+?)\.\.\.\.\.\.',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'dipole',
                                r'(POINT +\d+[\s\S]+?DEBYE.+\s+.+)',
                                repeats=True,
                                sub_parser=TextParser(quantities=[
                                    Quantity(
                                        'origin',
                                        rf'CHARGE\s+({re_f} +{re_f} +{re_f})',
                                        dtype=np.dtype(np.float64)
                                    ),
                                    Quantity(
                                        'value',
                                        rf'DEBYE\)\s+({re_f} +{re_f} +{re_f})',
                                        dtype=np.dtype(np.float64)
                                    )
                                ])
                            )
                        ])
                    )
                ])
            )
        ]

        self._quantities = [
            Quantity(
                'program_version',
                r'GAMESS VERSION *\= *(.+?) *\*|(Firefly version [\d\.]+)',
                str_operation=lambda x: x.strip()
            ),
            Quantity(
                'x_gamess_program_implementation',
                r'\*{16} (\d+[A-Z ]+) +\*{16}',
                str_operation=lambda x: x.strip()
            ),
            Quantity(
                'x_gamess_program_execution_date',
                r'EXECUTION OF GAMESS BEGUN ([\w\: ]+)',
                str_operation=lambda x: x.strip()
            ),
            Quantity(
                'x_gamess_memory',
                r'(\d+) WORDS OF MEMORY AVAILABLE', dtype=np.int32
            ),
            Quantity(
                'basis_options',
                rf'BASIS OPTIONS\s+\-+([\s\S]+?){re_n} *{re_n} *{re_n}',
                sub_parser=TextParser(quantities=[key_val_quantity])
            ),
            Quantity(
                'control_options',
                rf'CONTRL OPTIONS\s+\-+([\s\S]+?)\$',
                sub_parser=TextParser(quantities=[key_val_quantity])
            ),
            Quantity(
                'system_options',
                rf'SYSTEM OPTIONS\s+\-+([\s\S]+?)\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-',
                sub_parser=TextParser(quantities=[key_val_quantity])
            ),
            Quantity(
                'parameters',
                rf'(NUMBER OF BASIS SET SHELLS += +\d+[\s\S]+?){re_n} *{re_n}',
                str_operation=lambda x: [v.strip().split('=') for v in x.strip().splitlines()]
            ),
            Quantity(
                'geometry_opt',
                r'GEOMETRY SEARCH POINT([\s\S]+?)(?:BEGINNING|END OF GEOMETRY SEARCH|\Z)',
                repeats=True, sub_parser=TextParser(quantities=calc_quantities)
            )
        ] + calc_quantities


class GamessParser:
    def __init__(self):
        self._nspin = None
        self._units = {'ANGS': ureg.angstrom, 'BOHR': ureg.bohr}
        self._xc_map = {
            'SLATER': [{'name': 'LDA_X'}],
            'VWN': [{'name': 'LDA_C_VWN_5'}],
            'VWN3': [{'name': 'LDA_C_VWN_3'}],
            'VWN1RPA': [{'name': 'LDA_C_VWN1RPA'}],
            'BECKE': [{'name': 'GGA_X_B88'}],
            'OPTX': [{'name': 'GGA_X_OPTX'}],
            'GILL': [{'name': 'GGA_X_G96'}],
            'PW91X': [{'name': 'GGA_X_PW91'}],
            'PBEX': [{'name': 'GGA_X_PBE'}],
            'PZ81': [{'name': 'GGA_C_PZ'}],
            'P86': [{'name': 'GGA_C_P86'}],
            'LYP': [{'name': 'GGA_C_LYP'}],
            'PW91C': [{'name': 'GGA_C_PW91'}],
            'PBEC': [{'name': 'GGA_C_PBE'}],
            'OP': [{'name': 'GGA_C_OP'}],
            'SVWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'LDA_X'}],
            'SVWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'LDA_X'}],
            'SVWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'LDA_X'}],
            'SPZ81': [{'name': 'GGA_C_PZ'}, {'name': 'LDA_X'}],
            'SP86': [{'name': 'GGA_C_P86'}, {'name': 'LDA_X'}],
            'SLYP': [{'name': 'GGA_C_LYP'}, {'name': 'LDA_X'}],
            'SPW91': [{'name': 'GGA_C_PW91'}, {'name': 'LDA_X'}],
            'SPBE': [{'name': 'GGA_C_PBE'}, {'name': 'LDA_X'}],
            'SOP': [{'name': 'GGA_C_OP'}, {'name': 'LDA_X'}],
            'BVWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'LDA_X_B88'}],
            'BVWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'LDA_X_B88'}],
            'BVWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'LDA_X_B88'}],
            'BPZ81': [{'name': 'GGA_C_PZ'}, {'name': 'LDA_X_B88'}],
            'BP86': [{'name': 'GGA_C_P86'}, {'name': 'LDA_X_B88'}],
            'BLYP': [{'name': 'GGA_C_LYP'}, {'name': 'LDA_X_B88'}],
            'BPW91': [{'name': 'GGA_C_PW91'}, {'name': 'LDA_X_B88'}],
            'BPBE': [{'name': 'GGA_C_PBE'}, {'name': 'LDA_X_B88'}],
            'BOP': [{'name': 'GGA_C_OP'}, {'name': 'LDA_X_B88'}],
            'GVWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'GGA_X_G96'}],
            'GVWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'GGA_X_G96'}],
            'GVWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'GGA_X_G96'}],
            'GPZ81': [{'name': 'GGA_C_PZ'}, {'name': 'GGA_X_G96'}],
            'GP86': [{'name': 'GGA_C_P86'}, {'name': 'GGA_X_G96'}],
            'GLYP': [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_G96'}],
            'GPW91': [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_G96'}],
            'GPBE': [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_G96'}],
            'GOP': [{'name': 'GGA_C_OP'}, {'name': 'GGA_X_G96'}],
            'OVWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'GGA_X_OPTX'}],
            'OVWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'GGA_X_OPTX'}],
            'OVWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'GGA_X_OPTX'}],
            'OPZ81': [{'name': 'GGA_C_PZ'}, {'name': 'GGA_X_OPTX'}],
            'OP86': [{'name': 'GGA_C_P86'}, {'name': 'GGA_X_OPTX'}],
            'OLYP': [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_OPTX'}],
            'OPW91': [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_OPTX'}],
            'OPBE': [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_OPTX'}],
            'OOP': [{'name': 'GGA_C_OP'}, {'name': 'GGA_X_OPTX'}],
            'PW91VWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'GGA_X_PW91'}],
            'PW91VWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'GGA_X_PW91'}],
            'PW91VWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'GGA_X_PW91'}],
            'PW91PZ81': [{'name': 'GGA_C_PZ'}, {'name': 'GGA_X_PW91'}],
            'PW91P86': [{'name': 'GGA_C_P86'}, {'name': 'GGA_X_PW91'}],
            'PW91LYP': [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_PW91'}],
            'PW91': [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_PW91'}],
            'PW91PBE': [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PW91'}],
            'PW91OP': [{'name': 'GGA_C_OP'}, {'name': 'GGA_X_PW91'}],
            'PBEVWN': [{'name': 'LDA_C_VWN_5'}, {'name': 'GGA_X_PBE'}],
            'PBEVWN1RPA': [{'name': 'LDA_C_VWN1RPA'}, {'name': 'GGA_X_PBE'}],
            'PBEVWN3': [{'name': 'LDA_C_VWN_3'}, {'name': 'GGA_X_PBE'}],
            'PBEPZ81': [{'name': 'GGA_C_PZ'}, {'name': 'GGA_X_PBE'}],
            'PBEP86': [{'name': 'GGA_C_P86'}, {'name': 'GGA_X_PBE'}],
            'PBELYP': [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_PBE'}],
            'PBEPW91': [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_PBE'}],
            'PBE': [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PBE'}],
            'PBEOP': [{'name': 'GGA_C_OP'}, {'name': 'GGA_X_PBE'}],
            'EDF1': [{'name': 'GGA_XC_EDF1'}],
            'REVPBE': [{'name': 'GGA_XC_REVPBE'}],
            'RPBE': [{'name': 'GGA_XC_RPBE'}],
            'PBESOL': [{'name': 'GGA_XC_PBESOL'}],
            'HCTH93': [{'name': 'GGA_XC_HCTH_93'}],
            'HCTH147': [{'name': 'GGA_XC_HCTH_147'}],
            'HCTH407': [{'name': 'GGA_XC_HCTH_407'}],
            'SOGGA': [{'name': 'GGA_XC_SOGGA'}],
            'SOGGA11': [{'name': 'GGA_XC_SOGGA11'}],
            'MOHLYP': [{'name': 'GGA_XC_MOHLYP'}],
            'B97-D': [{'name': 'GGA_XC_B97D'}],
            'BHHLYP': [{'name': 'HYB_GGA_XC_BHANDHLYP'}],
            'B3PW91': [{'name': 'HYB_GGA_XC_B3PW91'}],
            'B3LYP': [{'name': 'HYB_GGA_XC_B3LYP'}],
            'B3LYPV1R': [{'name': 'HYB_GGA_XC_B3LYPVWN1RPA'}],
            'B3LYPV3': [{'name': 'HYB_GGA_XC_B3LYPVWN3'}],
            'B3P86': [{'name': 'HYB_GGA_XC_B3P86'}],
            'B3P86V1R': [{'name': 'HYB_GGA_XC_B3P86VWN1RPA'}],
            'B3P86V5': [{'name': 'HYB_GGA_XC_B3P86VWN5'}],
            'B97': [{'name': 'HYB_GGA_XC_B97'}],
            'B97-1': [{'name': 'HYB_GGA_XC_B971'}],
            'B97-2': [{'name': 'HYB_GGA_XC_B972'}],
            'B97-3': [{'name': 'HYB_GGA_XC_B973'}],
            'B97-K': [{'name': 'HYB_GGA_XC_B97K'}],
            'B98': [{'name': 'HYB_GGA_XC_B98'}],
            'PBE0': [{'name': 'HYB_GGA_XC_PBEH'}],
            'X3LYP': [{'name': 'HYB_GGA_XC_X3LYP'}],
            'SOGGA11X': [{'name': 'HYB_GGA_XC_SOGGA11X'}],
            'CAMB3LYP': [{'name': 'CAM-B3LYP'}],
            'WB97': [{'name': 'WB97'}],
            'WB97X': [{'name': 'WB97X'}],
            'WB97X-D': [{'name': 'WB97XD'}],
            'B2-PLYP': [{'name': 'B2PLYP'}],
            'B2K-PLYP': [{'name': 'B2KPLYP'}],
            'B2T-PLYP': [{'name': 'B2TPLYP'}],
            'B2GP-PLYP': [{'name': 'B2GPPLYP'}],
            'WB97X-2': [{'name': 'WB97X2'}],
            'WB97X-2L': [{'name': 'WB97X2L'}],
            'VS98': [{'name': 'MGGA_XC_VSXC'}],
            'PKZB': [{'name': 'MGGA_XC_PKZB'}],
            'THCTH': [{'name': 'MGGA_XC_TAU_HCTH'}],
            'THCTHHYB': [{'name': 'MGGA_XC_TAU_HCTHHYB'}],
            'BMK': [{'name': 'MGGA_XC_BMK'}],
            'TPSS': [{'name': 'MGGA_XC_TPSS'}],
            'TPSSH': [{'name': 'MGGA_XC_TPSSHYB'}],
            'TPSSM': [{'name': 'MGGA_XC_TPSSMOD'}],
            'REVTPSS': [{'name': 'MGGA_XC_REVISEDTPSS'}],
            'DLDF': [{'name': 'MGGA_XC_DLDF'}],
            'M05': [{'name': 'HYB_MGGA_XC_M05'}],
            'M05-2X': [{'name': 'HYB_MGGA_XC_M05_2X'}],
            'M06': [{'name': 'HYB_MGGA_XC_M06'}],
            'M06-L': [{'name': 'MGGA_C_M06_L'}, {'name': 'MGGA_X_M06_L'}],
            'M06-2X': [{'name': 'HYB_MGGA_XC_M06_2X'}],
            'M06-HF': [{'name': 'HYB_MGGA_XC_M06_HF'}],
            'M08-HX': [{'name': 'HYB_MGGA_XC_M08_HX'}],
            'M08-SO': [{'name': 'HYB_MGGA_XC_M08_SO'}],
            'M11-L': [{'name': 'MGGA_C_M11_L'}, {'name': 'MGGA_X_M11_L'}],
            'M11': [{'name': 'MGGA_C_M11_L'}, {'name': 'MGGA_X_M11'}],
            'RHF': [{'name': 'RHF_X'}],
            'UHF': [{'name': 'UHF_X'}],
            'ROHF': [{'name': 'ROHF_X'}],
            'LCBOPLRD': [{'name': 'HYB_GGA_XC_HSE03'}],
            'XALPHA': [{'name': 'XALPHA_X_GRIDFREE'}],
            'DEPRISTO': [{'name': 'DEPRISTO_X_GRIDFREE'}],
            'CAMA': [{'name': 'CAMA_X_GRIDFREE'}],
            'HALF': [{'name': 'HYB_GGA_XC_HALF_GRIDFREE'}],
            # 'VWN': [{'name': 'LDA_C_VWN5_GRIDFREE'}],
            'PWLOC': [{'name': 'PWLOC_C_GRIDFREE'}],
            'BPWLOC': [{'name': 'PWLOC_C_GRIDFREE'}, {'name': 'GGA_X_B88_GRIDFREE'}],
            'CAMB': [{'name': 'GGA_C_CAMBRIDGE_GRIDFREE'}, {'name': 'GGA_X_CAMA_GRIDFREE'}],
            'XVWN': [{'name': 'LDA_C_VWN5_GRIDFREE'}, {'name': 'XALPHA_X_GRIDFREE'}],
            'XPWLOC': [{'name': 'GGA_C_PW91_GRIDFREE'}, {'name': 'XALPHA_X_GRIDFREE'}],
            'SPWLOC': [{'name': 'PWLOC_C_GRIDFREE'}, {'name': 'LDA_X_GRIDFREE'}],
            'WIGNER': [{'name': 'GGA_XC_WIGNER_GRIDFREE'}],
            'WS': [{'name': 'GGA_XC_WIGNER_GRIDFREE'}],
            'WIGEXP': [{'name': 'GGA_XC_WIGNER_GRIDFREE'}],
            'NONE': [{'name': 'NONE'}],
        }
        self._method_map = {
            'RHF': [{'name': 'RHF'}],
            'UHF': [{'name': 'UHF'}],
            'ROHF': [{'name': 'ROHF'}],
            'GVB': [{'name': 'GVB'}],
            'MCSCF': [{'name': 'MCSCF'}],
            'EXCITE': [{'name': 'TDDFT'}],
            'SPNFLP': [{'name': 'SF-TDDFT'}],
            'POL': [{'name': 'HYPERPOL'}],
            'VB2000': [{'name': 'VB'}],
            'CIS': [{'name': 'CIS'}],
            'SFCIS': [{'name': 'SF-CIS'}],
            'ALDET': [{'name': 'DET-MCSCF'}],
            'ORMAS': [{'name': 'ORMAS'}],
            'FSOCI': [{'name': 'SECONDORDER-CI'}],
            'GENCI': [{'name': 'GENERAL-CI'}],
            'LCCD': [{'name': 'L-CCSD'}],
            'CCD': [{'name': 'CCD'}],
            'CCSD': [{'name': 'CCSD'}],
            'CCSD(T)': [{'name': 'CCSD(T)'}],
            'R-CC': [{'name': 'R-CCSD(T)&R-CCSD[T]'}],
            'CR-CC': [{'name': 'CR-CCSD(T)&CR-CCSD[T]'}],
            'CR-CCL': [{'name': 'CR-CC(2,3)'}],
            'CCSD(TQ)': [{'name': 'CCSD(TQ)&R-CCSD(TQ)'}],
            'CR-CC(Q)': [{'name': 'CR-CCSD(TQ)'}],
            'EOM-CCSD': [{'name': 'EOM-CCSD'}],
            'CR-EOM': [{'name': 'CR-EOMCCSD(T)'}],
            'CR-EOML': [{'name': 'CR-EOMCC(2,3)'}],
            'IP-EOM2': [{'name': 'IP-EOMCCSD'}],
            'IP-EOM3A': [{'name': 'IP-EOMCCSDt'}],
            'EA-EOM2': [{'name': 'EA-EOMCCSD'}],
            'EA-EOM3A': [{'name': 'EA-EOMCCSDt'}],
            'IOTC': [{'name': 'SCALAR_RELATIVISTIC'}],
            'DK': [{'name': 'SCALAR_RELATIVISTIC'}],
            'RESC': [{'name': 'SCALAR_RELATIVISTIC'}],
            'NESC': [{'name': 'SCALAR_RELATIVISTIC'}],
            'SBKJC': [{'name': 'PSEUDO_SCALAR_RELATIVISTIC'}],
            'HW': [{'name': 'PSEUDO_SCALAR_RELATIVISTIC'}],
            'MCP': [{'name': 'PSEUDO_SCALAR_RELATIVISTIC'}],
            'TAMMD': [{'name': 'TAMM-DANCOFF'}],
            'DFTB': [{'name': 'DFTB'}],
            'MP2': [{'name': 'MP2'}],
            'RIMP2': [{'name': 'RESOLUTIONOFIDENTITY-MP2'}],
            'CPHF': [{'name': 'COUPLEDPERTURBED-HF'}],
            'G3MP2': [{'name': 'G3(MP2)'}],
            'G32CCSD': [{'name': 'G3(MP2,CCSD(T))'}],
            'G4MP2': [{'name': 'G4(MP2)'}],
            'G4MP2-6X': [{'name': 'G4(MP2)-6X'}],
            'CCCA-S4': [{'name': 'CCCA-S4'}],
            'CCCA-CCL': [{'name': 'CCCA-CC(2,3)'}],
            'MCQDPT': [{'name': 'MCQDPT2'}],
            'DETMRPT': [{'name': 'MRPT2'}],
            'FORS': [{'name': 'CASSCF'}],
            'FOCI': [{'name': 'FIRSTORDER-CI'}],
            'SOCI': [{'name': 'SECONDORDER-CI'}],
            'DM': [{'name': 'TRANSITIONMOMENTS'}],
            'HSO1': [{'name': 'ONEELEC-SOC'}],
            'HSO2P': [{'name': 'PARTIALTWOELEC-SOC'}],
            'HSO2': [{'name': 'TWOELEC-SOC'}],
            'HSO2FF': [{'name': 'TWOELECFORMFACTOR-SOC'}],
            'GUGA': [{'name': 'GUGA-CI'}],
            'NONE': [{'name': 'NOCORRELATEDMETHOD'}],
        }
        self._basis_set_map = {
            'STO': [{'name': 'STO'}],
            'N21': [{'name': 'N21'}],
            'N31': [{'name': 'N31'}],
            'N311': [{'name': 'N311'}],
            'MINI': [{'name': 'MINI'}],
            'MIDI': [{'name': 'MIDI'}],
            'DZV': [{'name': 'DZV'}],
            'DH': [{'name': 'DH'}],
            'TZV': [{'name': 'TZV'}],
            'MC': [{'name': 'MC'}],
            'G3L': [{'name': 'G3MP2LARGE'}],
            'G3LX': [{'name': 'G3MP2LARGEXP'}],
            'CCD': [{'name': 'CC_PVDZ'}],
            'CCT': [{'name': 'CC_PVTZ'}],
            'CCQ': [{'name': 'CC_PVQZ'}],
            'CC5': [{'name': 'CC_PV5Z'}],
            'CC6': [{'name': 'CC_PV6Z'}],
            'ACCD': [{'name': 'AUG-CC-PVDZ'}],
            'ACCT': [{'name': 'AUG-CC-PVTZ'}],
            'ACCQ': [{'name': 'AUG-CC-PVQZ'}],
            'ACC5': [{'name': 'AUG-CC-PV5Z'}],
            'ACC6': [{'name': 'AUG-CC-PV6Z'}],
            'CCDC': [{'name': 'CC-PCVDZ'}],
            'CCTC': [{'name': 'CC-PCVTZ'}],
            'CCQC': [{'name': 'CC-PCVQZ'}],
            'CC5C': [{'name': 'CC-PCV5Z'}],
            'CC6C': [{'name': 'CC-PCV6Z'}],
            'ACCDC': [{'name': 'AUG-CC-PCVDZ'}],
            'ACCTC': [{'name': 'AUG-CC-PCVTZ'}],
            'ACCQC': [{'name': 'AUG-CC-PCVQZ'}],
            'ACC5C': [{'name': 'AUG-CC-PCV5Z'}],
            'ACC6C': [{'name': 'AUG-CC-PCV6Z'}],
            'CCDWC': [{'name': 'CC-PWCVDZ'}],
            'CCTWC': [{'name': 'CC-PWCVTZ'}],
            'CCQWC': [{'name': 'CC-PWCVQZ'}],
            'CC5WC': [{'name': 'CC-PWCV5Z'}],
            'CC6WC': [{'name': 'CC-PWCV6Z'}],
            'ACCDWC': [{'name': 'AUG-CC-PWCVDZ'}],
            'ACCTWC': [{'name': 'AUG-CC-PWCVTZ'}],
            'ACCQWC': [{'name': 'AUG-CC-PWCVQZ'}],
            'ACC5WC': [{'name': 'AUG-CC-PWCV5Z'}],
            'ACC6WC': [{'name': 'AUG-CC-PWCV6Z'}],
            'PCSEG-0': [{'name': 'PCSEG-0'}],
            'PCSEG-1': [{'name': 'PCSEG-1'}],
            'PCSEG-2': [{'name': 'PCSEG-2'}],
            'PCSEG-3': [{'name': 'PCSEG-3'}],
            'PCSEG-4': [{'name': 'PCSEG-4'}],
            'APCSEG-0': [{'name': 'AUG-PCSEG-0'}],
            'APCSEG-1': [{'name': 'AUG-PCSEG-1'}],
            'APCSEG-2': [{'name': 'AUG-PCSEG-2'}],
            'APCSEG-3': [{'name': 'AUG-PCSEG-3'}],
            'APCSEG-4': [{'name': 'AUG-PCSEG-4'}],
            'SPK-DZP': [{'name': 'SPK-DZP'}],
            'SPK-DTP': [{'name': 'SPK-DTP'}],
            'SPK-DQP': [{'name': 'SPK-DQP'}],
            'SPK-ADZP': [{'name': 'AUG-SPK-DZP'}],
            'SPK-ATZP': [{'name': 'AUG-SPK-TZP'}],
            'SPK-AQZP': [{'name': 'AUG-SPK-QZP'}],
            'SPKRDZP': [{'name': 'SPK-RELDZP'}],
            'SPKRDTP': [{'name': 'SPK-RELDTP'}],
            'SPKRDQP': [{'name': 'SPK-RELDQP'}],
            'SPKRADZP': [{'name': 'AUG-SPK-RELDZP'}],
            'SPKRATZP': [{'name': 'AUG-SPK-RELTZP'}],
            'SPKRAQZP': [{'name': 'AUG-SPK-RELQZP'}],
            'SPK-DZC': [{'name': 'SPK-DZC'}],
            'SPK-TZC': [{'name': 'SPK-TZC'}],
            'SPK-QZC': [{'name': 'SPK-QZC'}],
            'SPK-DZCD': [{'name': 'SPK-DZCD'}],
            'SPK-TZCD': [{'name': 'SPK-TZCD'}],
            'SPK-QZCD': [{'name': 'SPK-QZCD'}],
            'SPKRDZC': [{'name': 'SPK-RELDZC'}],
            'SPKRTZC': [{'name': 'SPK-RELTZC'}],
            'SPKRQZC': [{'name': 'SPK-RELQZC'}],
            'SPKRDZCD': [{'name': 'SPK-RELDZCD'}],
            'SPKRTZCD': [{'name': 'SPK-RELTZCD'}],
            'SPKRQZCD': [{'name': 'SPK-RELQZCD'}],
            'KTZV': [{'name': 'KARLSRUHETZV'}],
            'KTZVP': [{'name': 'KARLSRUHETZVP'}],
            'KTZVPP': [{'name': 'KARLSRUHETZVPP'}],
            'MCP-DZP': [{'name': 'MCP-DZP'}],
            'MCP-TZP': [{'name': 'MCP-TZP'}],
            'MCP-QZP': [{'name': 'MCP-QZP'}],
            'MCP-ATZP': [{'name': 'AUG-MCP-TZP'}],
            'MCP-AQZP': [{'name': 'AUG-MCP-QZP'}],
            'MCPCDZP': [{'name': 'MCPCDZP'}],
            'MCPCTZP': [{'name': 'MCPCTZP'}],
            'MCPCQZP': [{'name': 'MCPCQZP'}],
            'MCPACDZP': [{'name': 'AUG-MCPCDZP'}],
            'MCPACTZP': [{'name': 'AUG-MCPCTZP'}],
            'MCPACQZP': [{'name': 'AUG-MCPCQZP'}],
            'IMCP-SR1': [{'name': 'IMPROVEDMCP-SCALREL1'}],
            'IMCP-SR2': [{'name': 'IMPROVEDMCP-SCALREL2'}],
            'ZFK3-DK3': [{'name': 'ZFK3-DK3'}],
            'ZFK4-DK3': [{'name': 'ZFK4-DK3'}],
            'ZFK5-DK3': [{'name': 'ZFK5-DK3'}],
            'ZFK3LDK3': [{'name': 'ZFK3LDK3'}],
            'ZFK4LDK3': [{'name': 'ZFK4LDK3'}],
            'ZFK5LDK3': [{'name': 'ZFK5LDK3'}],
            'SLATER-MNDO': [{'name': 'SLATER-MNDO'}],
            'AM1': [{'name': 'SLATER-AM1'}],
            'PM3': [{'name': 'SLATER-PM3'}],
            'RM1': [{'name': 'SLATER-RM1'}],
            'DFTB': [{'name': 'SLATER-DFTB'}]
        }
        self._metainfo_map = {
            'one electron': 'x_gamess_one_electron',
            'two electron': 'x_gamess_two_electron',
            'virial ratio (v/t)': 'x_gamess_virial_ratio'
        }

        self.out_parser = OutParser()

    @property
    def nspin(self):
        if self._nspin is None:
            for key, val in self.out_parser.get('control_options', {}).get('key_val', []):
                if key == 'MULT':
                    self._nspin = 1 if val == 1 else 2
                    break
        return self._nspin

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self._nspin = None

    def parse_method(self):
        sec_method = self.archive.run[-1].m_create(Method)

        for key in ['basis', 'control', 'system']:
            setattr(sec_method, 'x_gamess_%s_options' % key, {
                key: val for key, val in self.out_parser.get('%s_options' % key, {}).get('key_val', [])})

        for key, val in self.out_parser.get('basis_options', {}).get('key_val', []):
            setattr(sec_method, 'x_gamess_basis_set_%s' % key.lower(), val)

        for value in self.out_parser.get('coordinates', {}).get('label_charge', []):
            atom = sec_method.m_create(AtomParameters)
            atom.label = value[0]
            atom.charge = float(value[1]) * ureg.elementary_charge

        def resolve_basis(gbasis):
            # basis_set_name = self._basis_set_map.get(gbasis)
            igauss = sec_method.x_gamess_basis_options.get('IGAUSS', 'NONE')

            if gbasis == 'STO':
                return f'STO-{igauss}G'

            symbol, igauss = ('G', f'{igauss}-') if gbasis.startswith('N') else ('', '')
            gbasis = gbasis.lstrip('N')

            plus = ''
            for orb in ['SP', 'P']:
                if sec_method.x_gamess_basis_options.get(f'DIFF{orb}', 'NONE').startswith('T'):
                    plus += '+'

            orb_suffix = ''
            for orb in ['d', 'f', 'p']:
                nfunc = sec_method.x_gamess_basis_options.get(f'N{orb.upper()}FUNC', 0)
                prefix = '' if nfunc == 0 else orb if nfunc == 1 else f'{nfunc}{orb}'
                orb_suffix += f',{prefix}' if orb == 'p' and orb_suffix and prefix else prefix
            orb_suffix = f'({orb_suffix})' if orb_suffix else orb_suffix

            return f'{igauss}{gbasis}{plus}{symbol}{orb_suffix}'

        gbasis = sec_method.x_gamess_basis_options.get('GBASIS', 'NONE')
        sec_basis = sec_method.m_create(BasisSet)
        sec_basis.name = resolve_basis(gbasis)
        for basis_set in self._basis_set_map.get(gbasis, []):
            sec_atom_basis = sec_basis.m_create(BasisSetAtomCentered)
            sec_atom_basis.name = basis_set.get('name')

        # electronic structure method
        sec_method.electronic = Electronic()
        relativity = sec_method.x_gamess_control_options.get(
            'RELWFN', sec_method.x_gamess_control_options.get('PP', 'NONE'))
        if relativity != 'NONE':
            for method in self._method_map.get(relativity, []):
                sec_method.electronic.relativity_method = method.get('name', '').lower()

        scf_type = sec_method.x_gamess_control_options.get('SCFTYP', 'NONE')
        run_type = sec_method.x_gamess_control_options.get('RUNTYP', 'NONE')
        mp_level = sec_method.x_gamess_control_options.get('MPLEVL', 'NONE')
        methods = [sec_method.x_gamess_control_options.get(
            key, 'NONE') for key in ['CITYP', 'CCTYP', 'VBTYP', 'TDDFT']]
        methods = [m for m in methods if m != 'NONE']
        electronic_method = scf_type if scf_type != 'NONE' else methods[0] if methods else run_type
        if scf_type in ['RHF', 'UHF', 'ROHF'] and mp_level == 2:
            sec_method.electronic.method = 'MP2'
        elif scf_type == 'MCSCF':
            sec_method.electronic.method = 'MRPT2' if mp_level == 2 else 'MCSCF'
        elif electronic_method != 'NONE':
            for method in self._method_map.get(electronic_method, []):
                sec_method.electronic.method = method.get('name')

        # dft xc functional
        dfttyp = sec_method.x_gamess_control_options.get('DFTTYP', 'NONE')
        if dfttyp != 'NONE':
            sec_method.dft = DFT(xc_functional=XCFunctional())
            for functional in self._xc_map.get(dfttyp, []):
                xc_functional = functional.get('name', '')
                if '_X_' in xc_functional or xc_functional.endswith('_X'):
                    sec_method.dft.xc_functional.exchange.append(Functional(name=xc_functional))
                elif '_C_' in xc_functional or xc_functional.endswith('_C'):
                    sec_method.dft.xc_functional.correlation.append(Functional(name=xc_functional))
                elif 'HYB' in xc_functional:
                    sec_method.dft.xc_functional.hybrid.append(Functional(name=xc_functional))
                else:
                    sec_method.dft.xc_functional.contributions.append(Functional(name=xc_functional))

    def parse_calculation(self, source):
        sec_scc = self.archive.run[-1].m_create(Calculation)
        if source.scf is not None:
            for iteration in source.scf.get('iteration', {}).get('iter', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                sec_scf.energy = Energy(
                    total=EnergyEntry(value=iteration[0] * ureg.hartree),
                    change=iteration[1] * ureg.hartree)

            sec_scc.energy = Energy(total=EnergyEntry(value=source.scf.energy_total))
            sec_scc.calculation_converged = source.scf.converged

            eigenvalues = source.scf.get('eigenvectors', {}).get('eigenvalues')
            if eigenvalues is not None:
                sec_eigenvalues = sec_scc.m_create(BandEnergies)
                eigenvalues = np.array(np.hstack(eigenvalues), dtype=np.float64)
                shape = (self.nspin, 1, np.size(eigenvalues) // self.nspin)
                sec_eigenvalues.energies = np.reshape(eigenvalues, shape) * ureg.hartree
                sec_eigenvalues.occupations = np.reshape(
                    np.array([2 if energy <= 0. else 0 for energy in eigenvalues]) / self.nspin, shape)

        properties = source.get('properties')
        if properties is not None:
            energies = {key.strip().lower(): val for key, val in properties.get('energy_components', [])}
            sec_energy = sec_scc.m_create(Energy)
            for key, val in energies.items():
                # if key.endswith('potential energy') or key.endswith('kinetic energy'):
                #     continue
                if key.endswith('energy'):
                    key = key.rstrip('energy').strip()
                    potential = energies.get(f'{key} potential energy')
                    # kinetic = energies.get(f'{key} kinetic energy')
                    key = self._metainfo_map.get(key, key).replace(' ', '_')
                    if hasattr(Energy, key) or key.startswith('x_gamess'):
                        setattr(sec_energy, key, EnergyEntry(
                            value=val * ureg.hartree,
                            potential=potential * ureg.hartree if potential else potential,
                            # kinetic=kinetic * ureg.hartree if kinetic else kinetic
                        ))
                    else:
                        sec_energy.contributions.append(EnergyEntry(kind=key, value=val * ureg.hartree))
                else:
                    setattr(sec_energy, self._metainfo_map.get(key, f'x_gamess_{key}'), val)

            population = properties.get('population_analysis')
            if population is not None:
                methods = ['Mulliken', 'Lowdin']
                for n, method in enumerate(methods):
                    sec_charges = sec_scc.m_create(Charges)
                    sec_charges.analysis_method = method
                    if population.atomic is not None:
                        sec_charges.value = population.atomic[n * 2 + 1] * ureg.elementary_charge
                    if population.atomic_orbitals is not None:
                        for norbital, orbital in enumerate(population.atomic_orbitals[1]):
                            sec_orbital_charges = sec_charges.m_create(ChargesValue, Charges.orbital_projected)
                            sec_orbital_charges.atom_index = population.atomic_orbitals[0][norbital]
                            sec_orbital_charges.orbital = str(orbital)
                            sec_orbital_charges.value = float(population.atomic_orbitals[2 + n][norbital]) * ureg.elementary_charge
                    if population.spins is not None:
                        sec_charges.spins = population.spins[n]

                orbitals = ['s', 'p', 'd', 'f', 'g', 'h', 'i']
                for nspin, charges in enumerate(population.get('spherical_harmonics', [])[:2]):
                    for norbital in range(len(charges[:-1])):
                        for natom in range(len(charges[norbital])):
                            sec_orbital_charges = sec_scc.charges[0].m_create(ChargesValue, Charges.orbital_projected)
                            sec_orbital_charges.atom_index = natom
                            sec_orbital_charges.spin = nspin
                            sec_orbital_charges.orbital = orbitals[norbital]
                            sec_orbital_charges.value = charges[norbital][natom] * ureg.elementary_charge

            moments = properties.get('electrostatic_moments')
            if moments is not None:
                for dipole in moments.get('dipole', []):
                    sec_multipoles = sec_scc.m_create(Multipoles)
                    sec_multipoles.kind = 'electrostatic'
                    sec_multipoles.dipole = MultipolesEntry(
                        # origin=dipole.get('origin'),
                        value=dipole.get('value'))

        if source.get('gradient') is not None:
            sec_scc.forces = Forces(total=ForcesEntry(value=source.gradient))

    def parse_system(self, source):
        coordinates = source.get('coordinates', self.out_parser.coordinates)
        if coordinates is not None:
            sec_system = self.archive.run[0].m_create(System)
            unit = self._units.get(coordinates.get('unit'), 1)
            sec_system.atoms = Atoms(
                labels=[chemical_symbols[int(lc[1])] for lc in coordinates.get('label_charge', [])],
                positions=np.array([p for p in coordinates.get('position', [])], dtype=np.float64) * unit)

    def parse(self, filepath, archive, logger):
        self.filepath = filepath
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.init_parser()

        run = archive.m_create(Run)
        run.program = Program(version=self.out_parser.get('program_version'))
        for key, val in self.out_parser.items():
            if key.startswith('x_gamess_'):
                setattr(run, key, val)

        self.parse_method()

        if self.out_parser.get('geometry_opt') is not None:
            for geometry_opt in self.out_parser.get('geometry_opt', []):
                self.parse_system(geometry_opt)
                self.parse_calculation(geometry_opt)
                run.calculation[-1].system_ref = run.system[-1]
                run.calculation[-1].method_ref = run.method[-1]
        else:
            self.parse_system(self.out_parser)
            self.parse_calculation(self.out_parser)
            run.calculation[-1].system_ref = run.system[-1]
            run.calculation[-1].method_ref = run.method[-1]

        # TODO
        # thermochemistry
        # parameters, output for diff methods e.g. mp2, cc, ci



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
import json
import numpy as np

from nomad.units import ureg
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.workflow import (
    Workflow, Elastic, EquationOfState, EOSFit, Thermodynamics, Stability, Decomposition,
    Phonon)
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.method import (
    Method, DFT, Electronic, XCFunctional, Functional, BasisSet, BasisSetCellDependent)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Dos, DosValues, BandStructure, BandEnergies)
from .metainfo.mp import Composition, Symmetry


class MPParser:
    def __init__(self):
        pass

    def init_parser(self):
        try:
            self.data = json.load(open(self.filepath))
        except Exception:
            self.logger.error('Failed to load json file.')

    def parse_elastic(self, source):
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'elastic'
        sec_elastic = sec_workflow.m_create(Elastic)
        sec_elastic.energy_stress_calculator = 'VASP'
        sec_elastic.calculation_method = 'stress'
        source = source.get('elasticity', source)
        sec_elastic.elastic_constants_order = source.get('order', 2)

        deformations = source.get('deformations')
        if deformations is not None:
            sec_elastic.n_deformations = len(deformations)

        elastic_tensor = source.get('elastic_tensor')
        if elastic_tensor is not None:
            sec_elastic.elastic_constants_matrix_second_order = elastic_tensor * ureg.GPa

        compliance_tensor = source.get('compliance_tensor')
        if compliance_tensor is not None:
            sec_elastic.compliance_matrix_second_order = compliance_tensor * (1 / ureg.GPa)

        if source.get('g_reuss') is not None:
            sec_elastic.shear_modulus_reuss = source['g_reuss'] * ureg.GPa
        if source.get('g_voigt') is not None:
            sec_elastic.shear_modulus_voigt = source['g_voigt'] * ureg.GPa
        if source.get('g_vrh') is not None:
            sec_elastic.shear_modulus_hill = source['g_vrh'] * ureg.GPa
        if source.get('homogeneous_poisson') is not None:
            sec_elastic.poisson_ratio_hill = source['homogeneous_poisson']
        if source.get('k_reuss') is not None:
            sec_elastic.bulk_modulus_reuss = source['k_reuss'] * ureg.GPa
        if source.get('k_voigt') is not None:
            sec_elastic.bulk_modulus_voigt = source['k_voigt'] * ureg.GPa
        if source.get('k_vrh') is not None:
            sec_elastic.bulk_modulus_hill = source['k_vrh'] * ureg.GPa

    def parse_eos(self, source):
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'equation_of_state'
        sec_eos = sec_workflow.m_create(EquationOfState)
        if source.get('volumes') is not None:
            sec_eos.volumes = source['volumes'] * ureg.angstrom ** 3
        if source.get('energies') is not None:
            sec_eos.energies = source['energies'] * ureg.eV
        for fit_function, result in source.get('eos', {}).items():
            sec_eos_fit = sec_eos.m_create(EOSFit)
            sec_eos_fit.function_name = fit_function
            if result.get('B') is not None:
                sec_eos_fit.bulk_modulus = result['B'] * ureg.eV / ureg.angstrom ** 3
            if result.get('C') is not None:
                sec_eos_fit.bulk_modulus_derivative = result['C']
            if result.get('E0') is not None:
                sec_eos_fit.equilibrium_energy = result['E0'] * ureg.eV
            if result.get('V0') is not None:
                sec_eos_fit.equilibrium_volume = result['V0'] * ureg.angstrom ** 3
            if result.get('eos_energies') is not None:
                sec_eos_fit.fitted_energies = result['eos_energies'] * ureg.eV

    def parse_thermo(self, data):
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'thermodynamics'
        sec_thermo = sec_workflow.m_create(Thermodynamics)
        sec_stability = sec_thermo.m_create(Stability)
        sec_stability.formation_energy = data.get(
            'formation_energy_per_atom', 0) * data.get('nsites', 1) * ureg.eV
        sec_stability.delta_formation_energy = data.get('energy_above_hull', 0) * ureg.eV
        sec_stability.is_stable = data.get('is_stable')
        if data.get('decomposes_to') is not None:
            for system in data.get('decomposes_to'):
                sec_decomposition = sec_stability.m_create(Decomposition)
                sec_decomposition.formula = system.get('formula')
                sec_decomposition.fraction = system.get('amount')

    def parse_phonon(self, data):
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'phonon'
        sec_phonon = sec_workflow.m_create(Phonon)
        # TODO is vasp always mp calculator?
        sec_phonon.force_calculator = 'vasp'

        calculations = self.archive.run[-1].calculation
        calc = calculations[-1] if calculations else self.archive.run[-1].m_create(Calculation)

        if data.get('ph_dos') is not None:
            sec_dos = calc.m_create(Dos, Calculation.dos_phonon)
            sec_dos.energies = data['ph_dos']['frequencies'] * ureg.THz * ureg.h
            dos = data['ph_dos']['densities'] * (1 / (ureg.THz * ureg.h))
            sec_dos.total.append(DosValues(value=dos))

        if data.get('ph_bs') is not None:
            sec_phonon.with_non_analytic_correction = data['ph_bs'].get('has_nac')
            sec_bs = calc.m_create(BandStructure, Calculation.band_structure_phonon)
            bands = np.transpose(data['ph_bs']['bands'])
            qpoints = data['ph_bs']['qpoints']
            labels = data['ph_bs']['labels_dict']
            hisym_qpts = list(labels.values())
            labels = list(labels.keys())
            endpoints = []
            for i, qpoint in enumerate(qpoints):
                if qpoint in hisym_qpts:
                    endpoints.append(i)
                if len(endpoints) < 2:
                    continue
                sec_segment = sec_bs.m_create(BandEnergies)
                energies = bands[endpoints[0]: endpoints[1] + 1]
                sec_segment.energies = np.reshape(energies, (1, *np.shape(energies))) * ureg.THz * ureg.h
                sec_segment.kpoints = qpoints[endpoints[0]: endpoints[1] + 1]
                sec_segment.endpoints_labels = [labels[hisym_qpts.index(qpoints[i])] for i in endpoints]
                endpoints = []

        calc.system_ref = self.archive.run[-1].system[0]

        # TODO add eigendisplacements

    def parse_tasks(self, data):
        if len(data['calcs_reversed']) == 0:
            return

        xc_func_mapping = {
            'PAW_PBE': ['GGA_X_PBE', 'GGA_C_PBE']
        }

        sec_method = self.archive.run[-1].m_create(Method)
        sec_xc_functional = XCFunctional()
        for potcar_type in data['calcs_reversed'][0].get('input', {}).get('potcar_type', []):
            for xc_functional in xc_func_mapping.get(potcar_type, []):
                if '_X_' in xc_functional or xc_functional.endswith('_X'):
                    sec_xc_functional.exchange.append(Functional(name=xc_functional))
                elif '_C_' in xc_functional or xc_functional.endswith('_C'):
                    sec_xc_functional.correlation.append(Functional(name=xc_functional))
                elif 'HYB' in xc_functional:
                    sec_xc_functional.hybrid.append(Functional(name=xc_functional))
                else:
                    sec_xc_functional.contributions.append(Functional(name=xc_functional))

        sec_method.dft = DFT(xc_functional=sec_xc_functional)
        sec_method.electronic = Electronic(method="DFT")

        encut = data['calcs_reversed'][0].get('input', {}).get('incar', {}).get('ENCUT')
        prec = data['calcs_reversed'][0].get('input', {}).get('incar', {}).get('PREC')
        if encut is not None and prec is not None:
            sec_basis = sec_method.m_create(BasisSet)
            sec_basis.type = 'plane waves'
            sec_basis_set_cell_dependent = sec_basis.m_create(BasisSetCellDependent)
            sec_basis_set_cell_dependent.kind = 'plane waves'
            prec = 1.3 if 'acc' in prec else 1.0
            sec_basis_set_cell_dependent.planewave_cutoff = encut * prec * ureg.eV

        self.archive.run[-1].calculation[0].method_ref = sec_method

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger(__name__)
        self.maindir = os.path.dirname(self.filepath)

        self.init_parser()

        sec_run = archive.m_create(Run)
        sec_run.program = Program(name='MaterialsProject', version="1.0.0")

        #  TODO system should be referenced
        structure = self.data.get('structure')
        if structure is not None:
            labels = [site['label'] for site in structure.get('sites')]
            positions = [site['xyz'] for site in structure.get('sites')]
            cell = structure.get('lattice', {}).get('matrix')
            sec_system = sec_run.m_create(System)
            sec_atoms = sec_system.m_create(Atoms)
            if cell is not None:
                sec_atoms.lattice_vectors = cell * ureg.angstrom
                sec_atoms.periodic = [True, True, True]
            if positions:
                sec_atoms.positions = positions * ureg.angstrom
            if labels:
                sec_atoms.labels = labels

        for key, val in self.data.get('composition', {}).items():
            sec_system.x_mp_composition.append(
                Composition(x_mp_label=key, x_mp_value=val))

        for key, val in self.data.get('composition_reduced', {}).items():
            sec_system.x_mp_composition_reduced.append(
                Composition(x_mp_label=key, x_mp_value=val))

        symmetry = self.data.get('symmetry')
        if symmetry is not None:
            sec_symmetry = sec_system.m_create(Symmetry)
            for key, val in symmetry.items():
                try:
                    setattr(sec_symmetry, 'x_mp_%s' % key, val)
                except Exception:
                    pass

        # misc
        sec_system.x_mp_elements = self.data.get('elements', [])
        for key, val in self.data.items():
            try:
                setattr(sec_system, 'x_mp_%s' % key, val)
            except Exception:
                pass

        # temporary fix to go through workflow normalization
        sec_calc = sec_run.m_create(Calculation)
        sec_calc.system_ref = sec_system

        # TODO should we use the MP api for workflow results?
        workflow_files = [f for f in os.listdir(
            self.maindir) if f.endswith('.json') and f != os.path.basename(self.filepath)]
        for filename in workflow_files:
            try:
                data = json.load(open(os.path.join(self.maindir, filename)))
            except Exception:
                continue
            # make sure data matches that of system
            # TODO maybe better to simply compare filename prefix so no need to load data
            if data.get('material_id', data.get('task_id')) != self.data.get('material_id'):
                continue

            if 'elasticity' in data:
                self.parse_elastic(data)
            if 'eos' in data:
                self.parse_eos(data)
            if 'ph_bs' in data or 'ph_dos' in data:
                self.parse_phonon(data)
            if 'property_name' in data and data.get('property_name') == 'thermo':
                self.parse_thermo(data)
            if 'calcs_reversed' in data:
                self.parse_tasks(data)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fastapi import FastAPI, status, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, RedirectResponse
import traceback

from nomad import config, utils

from .common import root_path
from .routers import dcat


logger = utils.get_logger(__name__)


app = FastAPI(
    root_path=root_path,
    openapi_url='/openapi.json',
    docs_url='/extensions/docs',
    redoc_url='/extensions/redoc',
    swagger_ui_oauth2_redirect_url='/extensions/docs/oauth2-redirect',
    title='DCAT API',
    version='v1, NOMAD %s@%s' % (config.meta.version, config.meta.commit),
    description='NOMAD\'s API for serving dcat resources')

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


async def redirect_to_docs(req: Request):
    return RedirectResponse(f'{root_path}/extensions/docs')


# app.add_route(f'{root_path}', redirect_to_docs, include_in_schema=False)
app.add_route('/', redirect_to_docs, include_in_schema=False)


@app.exception_handler(Exception)
async def unicorn_exception_handler(request: Request, e: Exception):
    logger.error('unexpected exception in API', url=request.url, exc_info=e)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            'detail': {
                'reason': 'Unexpected exception while handling your request',
                'exception': str(e),
                'exception_class': e.__class__.__name__,
                'exception_traceback': traceback.format_exc()
            }
        }
    )

app.include_router(dcat.router)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, MEnum, derived)
from nomad.datamodel.metainfo.simulation.system import System
from nomad.datamodel.metainfo.simulation.method import Method

from ..common import FastAccess


m_package = Package()


class ScfInfo(MCategory):
    '''
    Contains information on the self-consistent field (SCF) procedure, i.e. the number of
    SCF iterations (number_of_scf_iterations) or a section_scf_iteration section with
    detailed information on the SCF procedure of specified quantities.
    '''

    m_def = Category()


class AccessoryInfo(MCategory):
    '''
    Information that *in theory* should not affect the results of the calculations (e.g.,
    timing).
    '''

    m_def = Category()


class TimeInfo(MCategory):
    '''
    Stores information on the date and timings of the calculation. They are useful for,
    e.g., debugging or visualization purposes.
    '''

    m_def = Category(categories=[AccessoryInfo])


class EnergyValue(MCategory):
    '''
    This metadata stores an energy value.
    '''

    m_def = Category()


class EnergyTypeReference(MCategory):
    '''
    This metadata stores an energy used as reference point.
    '''

    m_def = Category(categories=[EnergyValue])


class ErrorEstimateContribution(MCategory):
    '''
    An estimate of a partial quantity contributing to the error for a given quantity.
    '''

    m_def = Category()


class Atomic(MSection):
    '''
    Generic section containing the values and information reqarding an atomic quantity
    such as charges, forces, multipoles.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of the quantity.
        ''')

    n_orbitals = Quantity(
        type=int,
        shape=[],
        description='''
        Number of orbitals used in the projection.
        ''')

    n_atoms = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms.
        ''')

    n_spin_channels = Quantity(
        type=int,
        shape=[],
        description='''
        Number of spin channels.
        ''')


class AtomicValues(MSection):
    '''
    Generic section containing information regarding the values of an atomic quantity.
    '''

    m_def = Section(validate=False)

    spin = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Spin channel corresponding to the atomic quantity.
        ''')

    atom_label = Quantity(
        type=str,
        shape=[],
        description='''
        Label of the atomic species corresponding to the atomic quantity.
        ''')

    atom_index = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Index of the atomic species corresponding to the atomic quantity.
        ''')

    m_kind = Quantity(
        type=str,
        shape=[],
        description='''
        String describing what the integer numbers of $m$ lm mean used in orbital
        projections. The allowed values are listed in the [m_kind wiki page]
        (https://gitlab.rzg.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/m-kind).
        ''')

    lm = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        Tuples of $l$ and $m$ values for which the atomic quantity are given. For
        the quantum number $l$ the conventional meaning of azimuthal quantum number is
        always adopted. For the integer number $m$, besides the conventional use as
        magnetic quantum number ($l+1$ integer values from $-l$ to $l$), a set of
        different conventions is accepted (see the [m_kind wiki
        page](https://gitlab.rzg.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/m-kind).
        The adopted convention is specified by m_kind.
        ''')

    orbital = Quantity(
        type=str,
        shape=[],
        description='''
        String representation of the of the atomic orbital.
        ''')


class EnergyEntry(Atomic):
    '''
    Section describing a type of energy or a contribution to the total energy.
    '''

    m_def = Section(validate=False)

    reference = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the reference energy to be subtracted from value to obtain a
        code-independent value of the energy.
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the energy of the unit cell.
        ''')

    value_per_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the energy normalized by the total number of atoms in the simulation
        cell.
        ''')

    values_per_atom = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms'],
        unit='joule',
        description='''
        Value of the atom-resolved energies.
        ''')

    potential = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the potential energy.
        ''')

    kinetic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the kinetic energy.
        ''')

    correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the correction to the energy.
        ''')


class Energy(MSection):
    '''
    Section containing all energy types and contributions.
    '''

    m_def = Section(validate=False)

    total = SubSection(
        sub_section=EnergyEntry.m_def,
        categories=[FastAccess],
        description='''
        Contains the value and information regarding the total energy of the system.
        ''')

    # TODO this should be removed and replaced by correction in EnergyEntry
    current = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the energy calculated with
        calculation_method_current. energy_current is equal to energy_total for
        non-perturbative methods. For perturbative methods, energy_current is equal to the
        correction: energy_total minus energy_total of the calculation_to_calculation_ref
        with calculation_to_calculation_kind = starting_point
        ''')

    zero_point = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the converged zero-point
        vibrations energy calculated using the method described in zero_point_method.
        ''')
    # this should be removed and replaced by electronic.kinetic
    kinetic_electronic = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the self-consistent electronic
        kinetic energy.
        ''')

    electronic = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the self-consistent electronic
        energy.
        ''')

    correlation = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the correlation energy calculated
        using the method described in XC_functional.
        ''')

    exchange = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the exchange energy calculated
        using the method described in XC_functional.
        ''')

    xc = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the exchange-correlation (XC)
        energy calculated with the functional stored in XC_functional.
        ''')

    # TODO Could potential be generalized for other energy types
    xc_potential = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the exchange-correlation (XC)
        potential energy: the integral of the first order derivative of the functional
        stored in XC_functional (integral of v_xc*electron_density), i.e., the component
        of XC that is in the sum of the eigenvalues. Value associated with the
        configuration, should be the most converged value..
        ''')

    electrostatic = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the total electrostatic energy
        (nuclei + electrons), defined consistently with calculation_method.
        ''')

    nuclear_repulsion = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the total nuclear-nuclear repulsion
        energy.
        ''')

    coulomb = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the Coulomb energy.
        ''')

    madelung = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the Madelung energy.
        ''')

    ewald = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the Ewald energy.
        ''')

    free = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the free energy (nuclei + electrons)
        (whose minimum gives the smeared occupation density calculated with
        smearing_kind).
        ''')

    sum_eigenvalues = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the sum of the eigenvalues of the
        Hamiltonian matrix.
        ''')

    total_t0 = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the total energy extrapolated to
        $T=0$, based on a free-electron gas argument.
        ''')

    van_der_waals = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains the value and information regarding the Van der Waals energy. A multiple
        occurence is expected when more than one van der Waals methods are defined. The
        van der Waals kind should be specified in Energy.kind
        ''')

    hartree_fock_x_scaled = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Scaled exact-exchange energy that depends on the mixing parameter of the
        functional. For example in hybrid functionals, the exchange energy is given as a
        linear combination of exact-energy and exchange energy of an approximate DFT
        functional; the exact exchange energy multiplied by the mixing coefficient of the
        hybrid functional would be stored in this metadata. Defined consistently with
        XC_method.
        ''')

    contributions = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains other energy contributions to the total energy not already defined.
        ''',
        repeats=True)

    types = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Contains other energy types not already defined.
        ''',
        repeats=True)

    # TODO remove this should be be entropy.correction
    correction_entropy = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Entropy correction to the potential energy to compensate for the change in
        occupation so that forces at finite T do not need to keep the change of occupation
        in account. Defined consistently with XC_method.
        ''')

    # TODO remove this should be in electrostatic.correction
    correction_hartree = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Correction to the density-density electrostatic energy in the sum of eigenvalues
        (that uses the mixed density on one side), and the fully consistent density-
        density electrostatic energy. Defined consistently with XC_method.
        ''')

    # TODO remove this should be in xc.correction
    correction_xc = SubSection(
        sub_section=EnergyEntry.m_def,
        description='''
        Correction to energy_XC.
        ''')

    change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Stores the change of total energy with respect to the previous step.
        ''',
        categories=[ErrorEstimateContribution, EnergyValue])

    fermi = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Fermi energy (separates occupied from unoccupied single-particle states)
        ''',
        categories=[EnergyTypeReference, EnergyValue])

    highest_occupied = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        The highest occupied energy.
        """)

    lowest_unoccupied = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        The lowest unoccupied energy.
        """)


class ForcesEntry(Atomic):
    '''
    Section describing a contribution to or type of atomic forces.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 3],
        unit='newton',
        description='''
        Value of the forces acting on the atoms. This is calculated as minus gradient of
        the corresponding energy type or contribution **including** constraints, if
        present. The derivatives with respect to displacements of nuclei are evaluated in
        Cartesian coordinates.  In addition, these are obtained by filtering out the
        unitary transformations (center-of-mass translations and rigid rotations for
        non-periodic systems, see value_raw for the unfiltered counterpart).
        ''')

    value_raw = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 3],
        unit='newton',
        description='''
        Value of the forces acting on the atoms **not including** such as fixed atoms,
        distances, angles, dihedrals, etc.''')


class Forces(MSection):
    '''
    Section containing all forces types and contributions.
    '''

    m_def = Section(validate=False)

    total = SubSection(
        sub_section=ForcesEntry.m_def,
        description='''
        Contains the value and information regarding the total forces on the atoms
        calculated as minus gradient of energy_total.
        ''')

    free = SubSection(
        sub_section=ForcesEntry.m_def,
        description='''
        Contains the value and information regarding the forces on the atoms
        corresponding to the minus gradient of energy_free. The (electronic) energy_free
        contains the information on the change in (fractional) occupation of the
        electronic eigenstates, which are accounted for in the derivatives, yielding a
        truly energy-conserved quantity.
        ''')

    t0 = SubSection(
        sub_section=ForcesEntry.m_def,
        description='''
        Contains the value and information regarding the forces on the atoms
        corresponding to the minus gradient of energy_T0.
        ''')

    contributions = SubSection(
        sub_section=ForcesEntry.m_def,
        description='''
        Contains other forces contributions to the total atomic forces not already
        defined.
        ''',
        repeats=True)

    types = SubSection(
        sub_section=ForcesEntry.m_def,
        description='''
        Contains other types of forces not already defined.
        ''',
        repeats=True)


class StressEntry(Atomic):
    '''
    Section describing a contribution to or a type of stress.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='joule/meter**3',
        description='''
        Value of the stress on the simulation cell. It is given as the functional
        derivative of the corresponding energy with respect to the deformation tensor.
        ''')

    values_per_atom = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3, 3],
        unit='joule/meter**3',
        description='''
        Value of the atom-resolved stresses.
        ''')


class Stress(MSection):
    '''
    Section containing all stress types and contributions.
    '''

    m_def = Section(validate=False)

    total = SubSection(
        sub_section=StressEntry.m_def,
        description='''
        Contains the value and information regarding the stress on the simulation cell
        and the atomic stresses corresponding to energy_total.
        ''')

    contributions = SubSection(
        sub_section=StressEntry.m_def,
        description='''
        Contains contributions for the total stress.
        ''',
        repeats=True)

    types = SubSection(
        sub_section=StressEntry.m_def,
        description='''
        Contains other types of stress.
        ''',
        repeats=True)


class ChargesValue(AtomicValues):
    '''
    Contains information on the charge on an atom or projected onto an orbital.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='coulomb',
        description='''
        Value of the charge projected on atom and orbital.
        ''')


class Charges(Atomic):
    '''
    Section describing the charges on the atoms obtained through a given analysis
    method. Also contains information on the orbital projection of charges.
    '''

    m_def = Section(validate=False)

    analysis_method = Quantity(
        type=str,
        shape=[],
        description='''
        Analysis method employed in evaluating the atom and partial charges.
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms'],
        unit='coulomb',
        description='''
        Value of the atomic charges calculated through analysis_method.
        ''')

    spins = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms'],
        description='''
        Value of the atomic spins.
        ''')

    total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='coulomb',
        description='''
        Value of the total charge of the system.
        ''')

    spin_projected = SubSection(sub_section=ChargesValue.m_def, repeats=True)

    orbital_projected = SubSection(sub_section=ChargesValue.m_def, repeats=True)


class BandGap(MSection):
    '''
    Band gap information for each spin channel.
    '''
    m_def = Section(
        description="""
        Contains information for each present spin channel.
        """)

    index = Quantity(
        type=np.dtype(np.int64),
        description="""
        The spin channel index.
        """)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Band gap value. Value of zero corresponds to not having a band gap.
        ''')

    type = Quantity(
        type=MEnum('direct', 'indirect'),
        shape=[],
        description='''
        Band gap type.
        ''')

    energy_highest_occupied = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        The highest occupied energy.
        """)

    energy_lowest_unoccupied = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        The lowest unoccupied energy.
        """)

    value_fundamental = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        GW fundamental band gap
        ''')

    value_optical = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        GW optical band gap
        ''')


class BandEnergies(MSection):
    '''
    This section describes the eigenvalue spectrum for a set of kpoints given by
    band_energies_kpoints.
    '''

    m_def = Section(validate=False)

    n_spin_channels = Quantity(
        type=int,
        shape=[],
        description='''
        Number of spin channels.
        ''')

    n_bands = Quantity(
        type=int,
        shape=[],
        description='''
        Number of bands for which the eigenvalues are evaluated.
        ''')

    n_kpoints = Quantity(
        type=int,
        shape=[],
        description='''
        Number of kpoints for which the eigenvalues are evaluated.
        ''')

    kpoints = Quantity(
        type=np.dtype(np.float64),
        shape=['n_kpoints', 3],
        description='''
        Fractional coordinates of the $k$ or $q$ points (in the basis of the reciprocal-
        lattice vectors) for which the eigenvalues are evaluated.
        ''')

    kpoints_weights = Quantity(
        type=np.dtype(np.float64),
        shape=['n_kpoints'],
        description='''
        Weights of the $k$ points in the calculation of the band energy.
        ''')

    kpoints_multiplicities = Quantity(
        type=np.dtype(np.float64),
        shape=['n_kpoints'],
        description='''
        Multiplicities of the $k$ point (i.e., how many distinct points per cell this
        expands to after applying all symmetries). This defaults to 1. If expansion is
        performed then each point will have weight
        band_energies_kpoints_weights/band_energies_kpoints_multiplicities.
        ''')

    endpoints_labels = Quantity(
        type=str,
        shape=[2],
        description='''
        Labels of the points along a one-dimensional path sampled in the $k$-space or
        $q$-space, using the conventional symbols, e.g., Gamma, K, L.
        ''')

    orbital_labels = Quantity(
        type=str,
        shape=['n_bands'],
        description='''
        Labels corresponding to each band/orbital
        ''')

    occupations = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        description='''
        Values of the occupations of the bands.
        ''')

    energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Values of the band energies.
        ''')

    qp_linearization_prefactor = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        description='''
        Values of the GW quasi particle linearization pre-factor.
        ''')

    value_xc_potential = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the GW exchange-correlation potential.
        ''')

    value_correlation = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the GW correlation energy.
        ''')

    value_exchange = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the GW exchange energy.
        ''')

    value_xc = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the GW exchange-correlation energy.
        ''')

    value_qp = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the GW quasi-particle energy.
        ''')

    value_ks = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the Kohn-Sham energy.
        ''')

    value_ks_xc = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        Diagonal matrix elements of the Kohn-Sham exchange-correlation energy.
        ''')

    band_gap = SubSection(sub_section=BandGap.m_def, repeats=True)


class BandStructure(MSection):
    '''
    This section stores information on a band structure evaluation along one-dimensional
    pathways in the $k$ or $q$ (reciprocal) space given in section_band_segment.
    Eigenvalues calculated at the actual $k$-mesh used for energy_total evaluations,
    can be found in the eigenvalues section.
    '''

    m_def = Section(validate=False)

    path_standard = Quantity(
        type=str,
        shape=[],
        description='''
        String to specify the standard used for the kpoints path within bravais
        lattice.
        ''')

    reciprocal_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='1 / meter',
        description='''
        The reciprocal cell within which the band structure is calculated.
        ''')

    band_gap = SubSection(sub_section=BandGap.m_def, repeats=True)

    energy_fermi = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        Fermi energy.
        """)

    segment = SubSection(sub_section=BandEnergies.m_def, repeats=True)


class DosFingerprint(MSection):
    '''
    Section for the fingerprint of the electronic density-of-states (DOS). DOS
    fingerprints are a modification of the D-Fingerprints reported in Chem. Mater. 2015,
    27, 3, 735–743 (doi:10.1021/cm503507h). The fingerprint consists of a binary
    representation of the DOS, that is used to evaluate the similarity of materials based
    on their electronic structure.
    '''

    m_def = Section(validate=False)

    bins = Quantity(
        type=str,
        shape=[],
        description='''
        Byte representation of the DOS fingerprint.
        ''')

    indices = Quantity(
        type=np.dtype(np.int16),
        shape=[2],
        description='''
        Indices used to compare DOS fingerprints of different energy ranges.
        ''')

    stepsize = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Stepsize of interpolation in the first step of the generation of DOS fingerprints.
        ''')

    filling_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Proportion of 1 bins in the DOS fingerprint.
        ''')

    grid_id = Quantity(
        type=str,
        shape=[],
        description='''
        Identifier of the DOS grid that was used for the creation of the fingerprint.
        Similarity can only be calculated if the same grid was used for both fingerprints.
        ''')


class DosValues(AtomicValues):
    '''
    Section containing information regarding the values of the density of states (DOS).
    '''

    m_def = Section(validate=False)

    phonon_mode = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon mode corresponding to the DOS used for phonon projections.
        ''')

    normalization_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / (m ** 3)',
        description='''
        Normalization factor for DOS values to get a cell-independent intensive DOS.
        For total dos, this is given by 1 / (unit cell volume).
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_energies'],
        unit='1/joule',
        description='''
        Values of DOS, i.e. number of states for a given energy. The set of discrete
        energy values is given in energies.
        ''')

    value_integrated = Quantity(
        type=np.dtype(np.float64),
        shape=['n_energies'],
        description='''
        Integrated DOS starting from the mimunum energy up to given a energy.
        ''')


class Dos(Atomic):
    '''
    Section containing information of an electronic-energy or phonon density of states
    (DOS) evaluation.
    '''

    m_def = Section(validate=False)

    n_energies = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of energy values for the DOS, see energies.
        ''')

    energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_energies'],
        unit='joule',
        description='''
        Contains the set of discrete energy values for the DOS.
        ''')

    energy_shift = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value necessary to shift the energies array so that the energy zero corresponds to
        the highest occupied energy level.
        ''')

    band_gap = SubSection(sub_section=BandGap.m_def, repeats=True)

    energy_fermi = Quantity(
        type=np.dtype(np.float64),
        unit="joule",
        shape=[],
        description="""
        Fermi energy.
        """)

    total = SubSection(sub_section=DosValues.m_def, repeats=True)

    atom_projected = SubSection(sub_section=DosValues.m_def, repeats=True)

    species_projected = SubSection(sub_section=DosValues.m_def, repeats=True)

    orbital_projected = SubSection(sub_section=DosValues.m_def, repeats=True)

    fingerprint = SubSection(sub_section=DosFingerprint.m_def, repeats=False)


class MultipolesValues(AtomicValues):
    '''
    Section containing the values of the multipoles projected unto an atom or orbital.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the multipole.
        ''')


class MultipolesEntry(Atomic):
    '''
    Section describing a multipole term. The unit of the values are given by C * m ^ n,
    where n = 1 for dipole, 2 for quadrupole, etc.
    '''

    m_def = Section(validate=False)

    origin = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        unit='meter',
        description='''
        Origin in cartesian space.
        ''')

    n_multipoles = Quantity(
        type=int,
        shape=[],
        description='''
        Number of multipoles.
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 'n_multipoles'],
        description='''
        Value of the multipoles projected unto the atoms.
        ''')

    total = Quantity(
        type=np.dtype(np.float64),
        shape=['n_multipoles'],
        description='''
        Total value of the multipoles.
        ''')

    orbital_projected = SubSection(sub_section=MultipolesValues.m_def, repeats=True)


class Multipoles(MSection):
    '''
    Section containing the multipoles (dipoles, quadrupoles, ...)
    for each atom.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of the multipoles being described.
        ''')

    dipole = SubSection(sub_section=MultipolesEntry.m_def, repeats=False)

    quadrupole = SubSection(sub_section=MultipolesEntry.m_def, repeats=False)

    octupole = SubSection(sub_section=MultipolesEntry.m_def, repeats=False)

    higher_order = SubSection(sub_section=MultipolesEntry.m_def, repeats=True)


class Thermodynamics(MSection):
    '''
    Section containing results related to a thermodynamics calculation.
    '''

    m_def = Section(validate=False)

    enthalpy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the calculated enthalpy per cell i.e. energy_total + pressure * volume.
        ''')

    entropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / kelvin',
        description='''
        Value of the entropy.
        ''')

    chemical_potential = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the chemical potential.
        ''')

    kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the kinetic energy.
        ''')

    potential_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the potential energy.
        ''')

    internal_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the internal energy.
        ''')

    vibrational_free_energy_at_constant_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the vibrational free energy per cell unit at constant volume.
        ''')

    pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Value of the pressure of the system.
        ''')

    temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        Value of the temperature of the system at which the properties are calculated.
        ''')

    volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='m ** 3',
        description='''
        Value of the volume of the system at which the properties are calculated.
        ''')

    heat_capacity_c_v = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / kelvin',
        description='''
        Stores the heat capacity per cell unit at constant volume.
        ''')

    heat_capacity_c_p = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / kelvin',
        description='''
        Stores the heat capacity per cell unit at constant pressure.
        ''')

    time_step = Quantity(
        type=int,
        shape=[],
        description='''
        The number of time steps with respect to the start of the calculation.
        ''')


class Volumetric(MSection):
    '''
    Section defining a set of volumetric data on a uniform real-space grid.
    Kind should be specified if the data is not explicitly defined by a metainfo class.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        The kind of function if not already defined.
        ''')

    multiplicity = Quantity(
        type=int,
        shape=[],
        description='''
        Number of functions stored.
        ''')

    n_x = Quantity(
        type=int,
        shape=[],
        description='''
        number of points along x axis
        ''')

    n_y = Quantity(
        type=int,
        shape=[],
        description='''
        number of points along y axis
        ''')

    n_z = Quantity(
        type=int,
        shape=[],
        description='''
        number of points along z axis
        ''')

    displacements = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='meter',
        description='''
        displacement vectors between grid points along each axis; same indexing rules as
        lattice_vectors.  In many cases, displacements and number of points are related to
        lattice_vectors through: [displacement] * [number of points + N] =
        [lattice_vector],where N is 1 for periodic directions and 0 for non-periodic ones
        ''')

    origin = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        unit='meter',
        description='''
        location of the first grid point; same coordinate system as atom_positions when
        applicable.
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['multiplicity', 'n_x', 'n_y', 'n_z'],
        description='''
        Values of the volumetric data defined by kind.
        ''')


class PotentialValue(Volumetric):
    '''
    Section containing the values of the potential evaluated on a uniform real-space grid.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['multiplicity', 'n_x', 'n_y', 'n_z'],
        unit='J / m ** 3',
        description='''
        Values of the potential evaluated at each grid point.
        ''')


class Potential(Volumetric):
    '''
    Section containing all potential types.
    '''

    m_def = Section(validate=False)

    effective = SubSection(sub_section=PotentialValue.m_def, repeats=True)

    hartree = SubSection(sub_section=PotentialValue.m_def, repeats=True)


class Density(Volumetric):
    '''
    Section containing the values of the density evaluated on a uniform real-space grid.
    '''

    m_def = Section(validate=False)

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['multiplicity', 'n_x', 'n_y', 'n_z'],
        unit='1 / m ** 3',
        description='''
        Values of the potential evaluated at each grid point.
        ''')


class ExcitedStates(MSection):
    '''
    Excited states properties.
    '''

    m_def = Section(validate=False)

    n_excited_states = Quantity(
        type=int,
        shape=[],
        description='''
        Number of excited states.
        ''')

    excitation_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_excited_states'],
        unit='joule',
        description='''
        Excitation energies.
        ''',
        categories=[EnergyValue])

    oscillator_strengths = Quantity(
        type=np.dtype(np.float64),
        shape=['n_excited_states'],
        description='''
        Excited states oscillator strengths.
        ''')

    transition_dipole_moments = Quantity(
        type=np.dtype(np.float64),
        shape=['n_excited_states', 3],
        unit='coulomb * meter',
        description='''
        Transition dipole moments.
        ''')


class VibrationsValues(MSection):
    '''
    Section describing a vibrational spectrum.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of the vibration.
        ''')

    description = Quantity(
        type=str,
        shape=[],
        description='''
        Short description of the vibration.
        ''')

    n_vibrations = Quantity(
        type=int,
        shape=[],
        description='''
        Number of values in the vibration spectrum.
        ''')

    activity = Quantity(
        type=str,
        shape=['n_vibrations'],
        description='''
        Describes the activity corresponding to each of the value of the vibration
        spectrum.
        ''')

    intensity = Quantity(
        type=np.dtype(np.float64),
        shape=['n_vibrations'],
        description='''
        Intensity of the vibration.
        ''')


class Vibrations(MSection):
    '''
    Section containing results related to vibrational frequencies.
    '''

    m_def = Section(validate=False)

    n_frequencies = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of vibration frequencies
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_frequencies'],
        unit='1 / meter',
        description='''
        Values of vibration Frequenices (cm-1)
        ''')

    raman = SubSection(sub_section=VibrationsValues.m_def, repeats=False)

    infrared = SubSection(sub_section=VibrationsValues.m_def, repeats=False)


class BaseCalculation(MSection):
    '''
    Contains computed properties of a configuration as defined by the corresponding
    section system and with the simulation method defined by section method. The
    references to the system and method sections are given by system_ref and method_ref,
    respectively.

    Properties derived from a group of configurations are not included in this section but
    can be accessed in section workflow.
    '''

    m_def = Section(validate=False)

    system_ref = Quantity(
        type=Reference(System.m_def),
        shape=[],
        description='''
        Links the calculation to a section system.
        ''',
        categories=[FastAccess])

    method_ref = Quantity(
        type=Reference(Method.m_def),
        shape=[],
        description='''
        Links the calculation to a section method.
        ''',
        categories=[FastAccess])

    starting_calculation_ref = Quantity(
        type=Reference(SectionProxy('Calculation')),
        shape=[],
        description='''
        Links the current section calculation to the starting calculation.
        ''',
        categories=[FastAccess])

    n_references = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
         Number of references to the current section calculation.
        ''')

    calculations_ref = Quantity(
        type=Reference(SectionProxy('Calculation')),
        shape=['n_references'],
        description='''
        Links the current section calculation to other section calculations. Such a link
        is necessary for example if the referenced calculation is a self-consistent
        calculation that serves as a starting point or a calculation is part of a domain
        decomposed simulation that needs to be connected.
        ''',
        categories=[FastAccess])

    calculations_path = Quantity(
        type=str,
        shape=['n_references'],
        description='''
        Links the current section calculation to other section calculations. Such a link
        is necessary for example if the referenced calculation is a self-consistent
        calculation that serves as a starting point or a calculation is part of a domain
        decomposed simulation that needs to be connected.
        ''')

    time_calculation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the wall-clock time needed for a calculation i.e. the real time that has
        been elapsed from start to end.
        ''',
        categories=[TimeInfo, AccessoryInfo])

    calculation_converged = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates whether a the calculation is converged.
        ''')

    hessian_matrix = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 'number_of_atoms', 3, 3],
        description='''
        The matrix with the second derivative of the energy with respect to atom
        displacements.
        ''')

    spin_S2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Stores the value of the total spin moment operator $S^2$ for the converged
        wavefunctions calculated with the XC_method. It can be used to calculate the spin
        contamination in spin-unrestricted calculations.
        ''')

    energy = SubSection(sub_section=Energy.m_def, categories=[FastAccess])

    forces = SubSection(sub_section=Forces.m_def)

    stress = SubSection(sub_section=Stress.m_def)

    dos_electronic = SubSection(sub_section=Dos.m_def, repeats=True)

    dos_phonon = SubSection(sub_section=Dos.m_def, repeats=True)

    eigenvalues = SubSection(sub_section=BandEnergies.m_def, repeats=True)

    band_structure_electronic = SubSection(sub_section=BandStructure.m_def, repeats=True)

    band_structure_phonon = SubSection(sub_section=BandStructure.m_def, repeats=True)

    thermodynamics = SubSection(sub_section=Thermodynamics.m_def, repeats=True)

    excited_states = SubSection(sub_section=ExcitedStates.m_def, repeats=True)

    vibrational_frequencies = SubSection(sub_section=Vibrations.m_def, repeats=True)

    potential = SubSection(sub_section=Potential.m_def, repeats=True)

    multipoles = SubSection(sub_section=Multipoles.m_def, repeats=True)

    charges = SubSection(sub_section=Charges.m_def, repeats=True)

    density_charge = SubSection(sub_section=Density.m_def, repeats=True)


class ScfIteration(BaseCalculation):
    '''
    Every scf_iteration section represents a self-consistent field (SCF) iteration,
    and gives detailed information on the SCF procedure of the specified quantities.
    '''

    m_def = Section(validate=False)


class Calculation(BaseCalculation):
    '''
    Every calculation section contains the values computed
    during a *single configuration calculation*, i.e. a calculation performed on a given
    configuration of the system (as defined in section_system) and a given computational
    method (e.g., exchange-correlation method, basis sets, as defined in section_method).

    The link between the current section calculation and the related
    system and method sections is established by the values stored in system_ref and
    method_ref, respectively.

    The reason why information on the system configuration and computational method is
    stored separately is that several *single configuration calculations* can be performed
    on the same system configuration, viz. several system configurations can be evaluated
    with the same computational method. This storage strategy avoids redundancies.
    '''

    m_def = Section(validate=False)

    n_scf_iterations = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of performed self-consistent field (SCF) iterations.
        ''',
        categories=[ScfInfo])

    scf_iteration = SubSection(sub_section=ScfIteration.m_def, repeats=True)


m_package.__init_metainfo__()
